from MyCapytain.common.base import Exportable
from MyCapytain.common.constants import get_graph, Mimetypes, RDF_NAMESPACES
from MyCapytain.errors import CitationDepthError
from copy import copy
from typing import Tuple
from abc import abstractmethod


class BaseCitationSet(Exportable):
    """ A citation set is a collection of citations that optionnaly can be matched using a .match() function

    :param children: List of Citation
    :type children: [BaseCitation]
    """

    def __repr__(self):
        return "<MyCapytain.common.reference.BaseCitationSet object at %s>" % id(self)

    EXPORT_TO = [Mimetypes.JSON.DTS.Std]

    def __init__(self, children=None):
        self._children = []

        if children:
            self.children = children

    @property
    def children(self):
        """ Children of a citation

        :rtype: [BaseCitation]
        """
        return self._children or []

    @children.setter
    def children(self, val: list):
        """ Sets children

        :param val: List of citation children
        """
        final_value = []
        if val is not None:
            for citation in val:
                if citation is None:
                    continue
                elif not isinstance(citation, (BaseCitation, type(self))):
                    raise TypeError("Citation children should be Citation")
                else:
                    if isinstance(self, BaseCitation):
                        citation.root = self.root
                    else:
                        citation.root = self
                    final_value.append(citation)

        self._children = final_value

    def add_child(self, child):
        """ Adds a child to the CitationSet

        :param child: Child citation to add
        :return:
        """
        if isinstance(child, BaseCitation):
            self._children.append(child)

    def __iter__(self):
        """ Iteration method

        Loop over the citation children

        :Example:
            >>>    c = BaseCitationSet(name="line")
            >>>    b = BaseCitationSet(name="poem", children=[c])
            >>>    b2 = BaseCitationSet(name="paragraph")
            >>>    a = BaseCitationSet(name="book", children=[b])
            >>>    [e for e in a] == [a, b, c, b2],

        """
        for child in self.children:
            yield from child

    @abstractmethod
    def match(self, passageId):
        """ Given a specific passageId, matches the citation to a specific citation object

        :param passageId: Passage Identifier
        :return: Citation that matches the passageId
        :rtype: BaseCitation
        """

    @property
    def depth(self):
        """ Depth of the citation scheme: if multiple scheme are available,
        the deepest one from the current node is chosen (if there is two possibilities,
        a->((b->c)(d->e->f)) 4 is the depth

        .. example:: If we have a Book, Poem, Line system,
         and the citation we are looking at is Poem, depth is 2


        :rtype: int
        :return: Depth of the citation scheme
        """
        if len(self.children):
            return max([child.depth for child in self.children])
        else:
            return 0

    def __getitem__(self, item):
        """ Returns the citations at the given level.

        :param item: Citation level
        :type item: int
        :rtype: list(BaseCitation) or BaseCitation

        """
        if item < 0:
            _item = self.depth + item
            if _item < 0:
                raise CitationDepthError("The negative %s is too small " % _item)
            item = _item
        if item == 0:
            yield from self.children
        else:
            for child in self.children:
                yield from child[item - 1]

    def __len__(self):
        """ Number of citation schemes covered by the object

        :rtype: int
        :returns: Number of nested citations
        """
        return len(self.children) + sum([len(child) for child in self.children])

    def __getstate__(self):
        """ Pickling method

        :return: dict
        """
        return copy(self.__dict__)

    def __setstate__(self, dic):
        self.__dict__ = dic
        return self

    def is_empty(self) -> bool:
        """ Check if the citation has not been set

        :return: True if nothing was setup
        :rtype: bool
        """
        return len(self.children) == 0

    def is_root(self) -> bool:
        """ Check if the Citation is the root

        :return:
        """
        return True

    def __export__(self, output=None, context=False, namespace_manager=None, **kwargs):
        if output == Mimetypes.JSON.DTS.Std:
            if not namespace_manager:
                namespace_manager = get_graph().namespace_manager

            _out = [
                cite.export(output, context=False, namespace_manager=namespace_manager)
                for cite in self.children
            ]

            if context:
                cite_structure_term = str(namespace_manager.qname(RDF_NAMESPACES.DTS.term("citeStructure")))
                _out = {
                    "@context": {
                        cite_structure_term.split(":")[0]: str(RDF_NAMESPACES.DTS)
                    },
                    cite_structure_term: _out
                }
            return _out


class BaseCitation(BaseCitationSet):
    def __repr__(self):
        """

        :return: String representation of the object
        :rtype: str
        """
        return "<{} name({})>".format(type(self).__name__, self.name)

    def __init__(self, name: str=None, children: list=None, root: BaseCitationSet=None):
        """ Initialize a BaseCitation object

        :param name: Name of the citation level
        :type name: str
        :param children: list of children
        :type children: [BaseCitation]
        :param root: Root of the citation group
        :type root: BaseCitationSet
        """
        super(BaseCitation, self).__init__()

        self._name = None
        self._root = None

        self.name = name
        self.children = children
        self.root = root

    def is_root(self) -> str:
        """
        :return: If the current object is the root of the citation set, True
        :rtype: bool
        """
        return self._root is None

    def is_set(self) -> bool:
        """ Checks that the current object is set

        :rtype: bool
        """
        return self.name is not None

    @property
    def root(self) -> BaseCitationSet:
        """ Returns the root of the citation set

        :return: Root of the Citation set
        :rtype: BaseCitationSet
        """
        if self._root is None:
            return self
        return self._root

    @root.setter
    def root(self, value):
        """ Set the root to which the current citation is connected to

        :param value: CitationSet root of the Citation graph
        :type value: BaseCitationSet
        :return:
        """
        self._root = value

    @property
    def name(self) -> str:
        """ Type of the citation represented

        :rtype: str
        :example: Book, Chapter, Textpart, Section, Poem...
        """
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    def __iter__(self):
        """ Iteration method

        Loop over the citation children

        :Example:
            >>>    c = BaseCitation(name="line")
            >>>    b = BaseCitation(name="poem", children=[c])
            >>>    b2 = BaseCitation(name="paragraph")
            >>>    a = BaseCitation(name="book", children=[b])
            >>>    [e for e in a] == [a, b, c, b2],

        """
        yield from [self]
        for child in self.children:
            yield from child

    def __export__(self, output=None, context=False, namespace_manager=None, **kwargs):
        if output == Mimetypes.JSON.DTS.Std:
            if not namespace_manager:
                namespace_manager = get_graph().namespace_manager

            cite_type_term = str(namespace_manager.qname(RDF_NAMESPACES.DTS.term("citeType")))
            cite_structure_term = str(namespace_manager.qname(RDF_NAMESPACES.DTS.term("citeStructure")))

            _out = {
                cite_type_term: self.name
            }

            if not self.is_empty():
                _out[cite_structure_term] = [
                    cite.export(output, context=False, namespace_manager=namespace_manager)
                    for cite in self.children
                ]

            if context:
                _out["@context"] = {cite_type_term.split(":")[0]: str(RDF_NAMESPACES.DTS)}

            return _out

    @property
    def depth(self) -> int:
        """ Depth of the citation scheme

        .. example:: If we have a Book, Poem, Line system, and the citation we are looking at is Poem, depth is 1


        :rtype: int
        :return: Depth of the citation scheme
        """
        if len(self.children):
            return 1 + max([child.depth for child in self.children])
        else:
            return 1


class BaseReference(tuple):
    """ BaseReference represents a passage identifier, range or not

    It is made of two major properties : .start and .end

    To check if the object is a range, you can use the method .is_range()
    """
    def __new__(cls, *refs):
        if len(refs) == 1 and not isinstance(refs[0], tuple):
            refs = refs[0], None

        obj = tuple.__new__(cls, refs)

        return obj

    def is_range(self) -> int:
        return bool(self[1])

    @property
    def start(self):
        """ Quick access property for start part

        :rtype: str
        """
        return self[0]

    @property
    def end(self):
        """ Quick access property for reference end list

        :rtype: str
        """
        return self[1]


class BaseReferenceSet(tuple):
    """ A BaseReferenceSet is a set of Reference (= a bag of identifier)
    that can carry citation and level information (what kind of reference is this reference ?
    Where am I in the levels of the current document ?)

    It can be iterate like a tuple and has a .citation and .level property

    """
    def __new__(cls, *refs, citation: BaseCitationSet=None, level: int=1):
        if len(refs) == 1 and not isinstance(refs, BaseReference):
            refs = refs[0]
        obj = tuple.__new__(cls, refs)
        obj._citation = None
        obj._level = level

        if citation is not None:
            obj._citation = citation
        return obj

    @property
    def citation(self) -> BaseCitationSet:
        return self._citation

    @property
    def level(self) -> int:
        """ Level represents the depth at which lies the reference.

        eg. depth is the number of floors, level is the actual floor number.
        """
        return self._level

    def __repr__(self):
        return "<{typ} ({repr}) level:{level}, citation:{citation}>".format(
            typ=type(self).__name__,
            repr=", ".join([str(s) for s in self]),
            level=self.level,
            citation=str(self.citation)
        )


class NodeId(object):
    """ Collection of directional references for a Tree

    :param identifier: Current object identifier
    :type identifier: str
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int
    """
    def __init__(self, identifier=None, children=None, parent=None, siblings=(None, None), depth=None):
        self._children = children or []
        self._parent = parent
        self._prev_id, self._next_id = siblings
        self._identifier = identifier
        self._depth = depth

    @property
    def depth(self) -> int:
        """ Depth of the node in the global hierarchy of the text tree
        """
        return self._depth

    @property
    def childIds(self) -> BaseReferenceSet:
        """ Children Ids
        """
        return self._children

    @property
    def firstId(self) -> BaseReference:
        """ First child Id
        """
        if len(self._children) == 0:
            return None
        return self._children[0]

    @property
    def lastId(self) -> BaseReference:
        """ Last child id
        """
        if len(self._children) == 0:
            return None
        return self._children[-1]

    @property
    def parentId(self) -> BaseReference:
        """ Parent Id
        """
        return self._parent

    @property
    def siblingsId(self) -> Tuple[BaseReference, BaseReference]:
        """ Siblings Id
        """
        return self.prevId, self.nextId

    @property
    def prevId(self) -> BaseReference:
        """ Previous Id (Sibling)
        """
        return self._prev_id

    @property
    def nextId(self) -> BaseReference:
        """ Next Id
        """
        return self._next_id

    @property
    def id(self):
        """Current object identifier

        :rtype: str
        """
        return self._identifier



from inspect import getmro


__all__ = [
    "Exportable"
]


class Exportable(object):
    """ Objects that supports Export

    :cvar EXPORT_TO: List of Mimetypes the resource can export to
    """
    EXPORT_TO = []
    DEFAULT_EXPORT = None

    def __init__(self, *args, **kwargs):
        pass

    @property
    def export_capacities(self):
        """  List Mimetypes that current object can export to
        """
        return [export for cls in getmro(type(self)) if hasattr(cls, "EXPORT_TO") for export in cls.EXPORT_TO]

    def __export__(self, output=None, **kwargs):
        """ Export the collection item in the Mimetype required.

        :param output: Mimetype to export to (Uses MyCapytain.common.utils.Mimetypes)
        :type output: str
        :return: Object using a different representation
        """
        return None

    def export(self, output=None, **kwargs):
        """ Export the collection item in the Mimetype required.

        :param output: Mimetype to export to (Uses MyCapytain.common.utils.Mimetypes)
        :type output: str
        :return: Object using a different representation
        """
        if output is None:
            output = self.DEFAULT_EXPORT
        if output is not None and output in self.export_capacities:
            for cls in getmro(type(self)):
                if hasattr(cls, "EXPORT_TO") and output in cls.EXPORT_TO:
                    return cls.__export__(self, output, **kwargs)
        raise NotImplementedError(
            "Mimetype {} has not been implemented for this resource".format(output or "(No Mimetype)")
        )


from copy import copy
from io import StringIO, IOBase
from xml.sax.saxutils import escape

from lxml.etree import SubElement
import lxml.etree as etree
from lxml.objectify import ObjectifiedElement, parse, Element
from six import text_type

from MyCapytain.common.constants import XPATH_NAMESPACES


__all__ = [
    "make_xml_node",
    "xmlparser",
    "normalizeXpath",
    "xmliter",
    "performXpath",
    "copyNode",
    "passageLoop"
]


_parser = etree.XMLParser(collect_ids=False, resolve_entities=False)


def make_xml_node(graph, name, close=False, attributes=None, text="", complete=False, innerXML=""):
    """ Create an XML Node

    :param graph: Graph used to geneates prefixes
    :param name: Name of the tag
    :param close: Produce closing tag (close=False -> "<tag>", close=True -> "</tag>")
    :param attributes: Dictionary of attributes
    :param text: CapitainsCtsText to put inside the node
    :param complete: Complete node (node with opening and closing tag)
    :param innerXML: XML to append to the node
    :return: String representation of the node
    :rtype: str
    """
    name = graph.namespace_manager.qname(name)
    if complete:
        if attributes is not None:
            return "<{0} {1}>{2}{3}</{0}>".format(
                name,
                " ".join(
                    [
                        "{}=\"{}\"".format(attr_name, attr_value)
                        for attr_name, attr_value in attributes.items()
                    ]
                ),
                escape(text),
                innerXML
            )
        return "<{0}>{1}{2}</{0}>".format(name, escape(text), innerXML)
    elif close is True:
        return "</{}>".format(name)
    elif attributes is not None:
        return "<{} {}>".format(
            name,
            " ".join(
                [
                    "{}=\"{}\"".format(attr_name, attr_value)
                    for attr_name, attr_value in attributes.items()
                ]
            )
        )
    return "<{}>".format(name)


def xmliter(node):
    """ Provides a simple XML Iter method which complies with either _Element or _ObjectifiedElement

    :param node: XML Node
    :return: Iterator for iterating over children of said node.
    """
    if hasattr(node, "iterchildren"):
        return node.iterchildren()
    else:
        return node


def xmlparser(xml, objectify=True):
    """ Parse xml

    :param xml: XML element
    :type xml: Union[text_type, lxml.etree._Element]
    :rtype: lxml.etree._Element
    :returns: An element object
    :raises: TypeError if element is not in accepted type

    """
    doclose = None
    if isinstance(xml, (etree._Element, ObjectifiedElement, etree._ElementTree)):
        return xml
    elif isinstance(xml, text_type):
        xml = StringIO(xml)
        doclose = True
    elif not isinstance(xml, IOBase):
        raise TypeError("Unsupported type of resource {}".format(type(xml)))

    if objectify is False:
        parsed = etree.parse(xml).getroot()
    else:
        parsed = parse(xml).getroot()
    if doclose:
        xml.close()
    return parsed


def __formatXpath__(xpath):
    """ Format at XPath for perform XPath

    :param xpath: XPath element lists
    :return: Tuple where the first element is an XPath representing the next node to retrieve and the second the list \
    of other elements to find
    """
    if len(xpath) > 1:
        current, queue = xpath[0], xpath[1:]
        current = "./{}[./{}]".format(
            current,
            "/".join(queue)
        )
    else:
        current, queue = "./{}".format(xpath[0]), []

    return current, queue


def performXpath(parent, xpath):
    """ Perform an XPath on an element and indicate if we need to loop over it to find something

    :param parent: XML Node on which to perform XPath
    :param xpath: XPath to run
    :return: (Result, Need to loop Indicator)
    """
    loop = False
    if xpath.startswith(".//"):
        result = parent.xpath(
            xpath.replace(".//", "./", 1),
            namespaces=XPATH_NAMESPACES
        )
        if len(result) == 0:
            result = parent.xpath(
                "*[{}]".format(xpath),
                namespaces=XPATH_NAMESPACES
            )
            loop = True
    else:
        result = parent.xpath(
            xpath,
            namespaces=XPATH_NAMESPACES
        )
    return result[0], loop


def copyNode(node, children=False, parent=False):
    """ Copy an XML Node

    :param node: Etree Node
    :param children: Copy children nodes is set to True
    :param parent: Append copied node to parent if given
    :return: New Element
    """
    if parent is not False:
        element = SubElement(
            parent,
            node.tag,
            attrib=node.attrib,
            nsmap={None: "http://www.tei-c.org/ns/1.0"}
        )
    else:
        element = Element(
            node.tag,
            attrib=node.attrib,
            nsmap={None: "http://www.tei-c.org/ns/1.0"}
        )
    if children:
        if node.text:
            element._setText(node.text)
        for child in xmliter(node):
            element.append(copy(child))
    return element


def normalizeXpath(xpath):
    """ Normalize XPATH split around slashes

    :param xpath: List of xpath elements
    :type xpath: [str]
    :return: List of refined xpath
    :rtype: [str]
    """
    new_xpath = []
    for x in range(0, len(xpath)):
        if x > 0 and len(xpath[x-1]) == 0:
            new_xpath.append("/"+xpath[x])
        elif len(xpath[x]) > 0:
            new_xpath.append(xpath[x])
    return new_xpath


def passageLoop(parent, new_tree, xpath1, xpath2=None, preceding_siblings=False, following_siblings=False):
    """ Loop over passages to construct and increment new tree given a parent and XPaths

    :param parent: Parent on which to perform xpath
    :param new_tree: Parent on which to add nodes
    :param xpath1: List of xpath elements
    :type xpath1: [str]
    :param xpath2: List of xpath elements
    :type xpath2: [str]
    :param preceding_siblings: Append preceding siblings of XPath 1/2 match to the tree
    :param following_siblings: Append following siblings of XPath 1/2 match to the tree
    :return: Newly incremented tree
    """
    current_1, queue_1 = __formatXpath__(xpath1)
    if xpath2 is None:  # In case we need what is following or preceding our node
        result_1, loop = performXpath(parent, current_1)
        if loop is True:
            queue_1 = xpath1

        central = None
        has_no_queue = len(queue_1) == 0
        # For each sibling, when we need them in the context of a range
        if preceding_siblings or following_siblings:
            for sibling in xmliter(parent):
                if sibling == result_1:
                    central = True
                    # We copy the node we looked for (Result_1)
                    child = copyNode(result_1, children=has_no_queue, parent=new_tree)
                    # if we don't have children
                    # we loop over the passage child
                    if not has_no_queue:
                        passageLoop(
                            result_1,
                            child,
                            queue_1,
                            None,
                            preceding_siblings=preceding_siblings,
                            following_siblings=following_siblings
                        )
                    # If we were waiting for preceding_siblings, we break it off
                    # As we don't need to go further
                    if preceding_siblings:
                        break
                elif not central and preceding_siblings:
                    copyNode(sibling, parent=new_tree, children=True)
                elif central and following_siblings:
                    copyNode(sibling, parent=new_tree, children=True)
    else:
        result_1, loop = performXpath(parent, current_1)
        if loop is True:
            queue_1 = xpath1
            if xpath2 == xpath1:
                current_2, queue_2 = current_1, queue_1
            else:
                current_2, queue_2 = __formatXpath__(xpath2)
        else:
            current_2, queue_2 = __formatXpath__(xpath2)

        if xpath1 != xpath2:
            result_2, loop = performXpath(parent, current_2)
            if loop is True:
                queue_2 = xpath2
        else:
            result_2 = result_1

        if result_1 == result_2:
            has_no_queue = len(queue_1) == 0
            child = copyNode(result_1, children=has_no_queue, parent=new_tree)
            if not has_no_queue:
                passageLoop(
                    result_1,
                    child,
                    queue_1,
                    queue_2
                )
        else:
            start = False
            # For each sibling
            for sibling in xmliter(parent):
                # If we have found start
                # We copy the node because we are between start and end
                if start:
                    # If we are at the end
                    # We break the copy
                    if sibling == result_2:
                        break
                    else:
                        copyNode(sibling, parent=new_tree, children=True)
                # If this is start
                # Then we copy it and initiate star
                elif sibling == result_1:
                    start = True
                    has_no_queue_1 = len(queue_1) == 0
                    node = copyNode(sibling, children=has_no_queue_1, parent=new_tree)
                    if not has_no_queue_1:
                        passageLoop(sibling, node, queue_1, None, following_siblings=True)

            continue_loop = len(queue_2) == 0
            node = copyNode(result_2, children=continue_loop, parent=new_tree)
            if not continue_loop:
                passageLoop(result_2, node, queue_2, None, preceding_siblings=True)

    return new_tree



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.texts.local.tei

This module contains methods to parse local resources using TEI/Epidoc guidelines of CapiTainS

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

import warnings

from typing import Tuple, Optional
from MyCapytain.errors import DuplicateReference, MissingAttribute, RefsDeclError, EmptyReference, CitationDepthError, MissingRefsDecl
from MyCapytain.common.utils.xml import copyNode, normalizeXpath, passageLoop
from MyCapytain.common.constants import XPATH_NAMESPACES
from MyCapytain.common.reference import CtsReference, URN, Citation, CtsReferenceSet

from MyCapytain.resources.prototypes.cts.text import PrototypeCtsPassage, PrototypeCtsText
from MyCapytain.resources.texts.base.tei import TeiResource

from MyCapytain.errors import InvalidSiblingRequest, InvalidURN
import lxml.etree as etree


__all__ = [
    "CapitainsCtsText",
    "CapitainsCtsPassage"
]


def _make_passage_kwargs(urn, reference):
    """ Little helper used by CapitainsCtsPassage here to comply with parents args

    :param urn: URN String
    :param reference: Reference String
    :return: Dictionary of arguments with URN based on identifier and reference
    """
    kwargs = {}
    if urn is not None:
        if reference is not None:
            kwargs["urn"] = URN("{}:{}".format(urn.upTo(URN.VERSION), reference))
        else:
            kwargs["urn"] = urn
    return kwargs


class _SharedMethods(TeiResource):
    """ Set of shared methods between objects in local TEI. Avoid recoding functions
    """

    def getTextualNode(self, subreference=None, simple=False):
        """ Finds a passage in the current text

        :param subreference: Identifier of the subreference / passages
        :type subreference: Union[list, CtsReference]
        :param simple: If set to true, retrieves nodes up to the given one, cleaning non required siblings.
        :type simple: boolean
        :rtype: CapitainsCtsPassage, ContextPassage
        :returns: Asked passage
        """
        if subreference is None:
            return self._getSimplePassage()

        if not isinstance(subreference, CtsReference):
            if isinstance(subreference, str):
                subreference = CtsReference(subreference)
            elif isinstance(subreference, list):
                subreference = CtsReference(".".join(subreference))

        if len(subreference.start) > self.citation.root.depth:
            raise CitationDepthError("URN is deeper than citation scheme")

        if simple is True:
            return self._getSimplePassage(subreference)

        if not subreference.is_range():
            start = end = subreference.start.list
        else:
            start, end = subreference.start.list, subreference.end.list

        citation_start = self.citation.root[len(start)-1]
        citation_end = self.citation.root[len(end)-1]

        start, end = citation_start.fill(passage=start), citation_end.fill(passage=end)
        start, end = normalizeXpath(start.split("/")[2:]), normalizeXpath(end.split("/")[2:])

        xml = self.textObject.xml

        if isinstance(xml, etree._Element):
            root = copyNode(xml)
        else:
            root = copyNode(xml.getroot())

        root = passageLoop(xml, root, start, end)

        if self.urn:
            urn = URN("{}:{}".format(self.urn, subreference))
        else:
            urn = None

        return CapitainsCtsPassage(
            urn=urn,
            resource=root,
            text=self,
            citation=citation_start,
            reference=subreference
        )

    def _getSimplePassage(self, reference=None):
        """ Retrieve a single node representing the passage.

        .. warning:: Range support is awkward.

        :param reference: Identifier of the subreference / passages
        :type reference: list, reference
        :returns: Asked passage
        :rtype: CapitainsCtsPassage
        """
        if reference is None:
            return _SimplePassage(
                resource=self.resource,
                reference=None,
                urn=self.urn,
                citation=self.citation.root,
                text=self
            )

        subcitation = self.citation.root[reference.depth-1]
        resource = self.resource.xpath(
            subcitation.fill(reference),
            namespaces=XPATH_NAMESPACES
        )

        if len(resource) != 1:
            raise InvalidURN

        return _SimplePassage(
            resource[0],
            reference=reference,
            urn=self.urn,
            citation=subcitation,
            text=self.textObject
        )

    @property
    def textObject(self):
        """ Textual Object with full capacities (Unlike Simple CapitainsCtsPassage)

        :rtype: CtsTextMetadata, CapitainsCtsPassage
        :return: Textual Object with full capacities (Unlike Simple CapitainsCtsPassage)
        """
        text = None
        if isinstance(self, CapitainsCtsText):
            text = self
        return text

    def getReffs(self, level: int=1, subreference: CtsReference=None) -> CtsReferenceSet:
        """ CtsReference available at a given level

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :param subreference: Subreference (optional)
        :returns: List of levels
        """

        if not subreference and hasattr(self, "reference"):
            subreference = self.reference
        elif subreference and not isinstance(subreference, CtsReference):
            subreference = CtsReference(subreference)

        return self.getValidReff(level=level, reference=subreference)

    def getValidReff(self, level: int=1, reference: CtsReference=None, _debug: bool=False) -> CtsReferenceSet:
        """ Retrieve valid passages directly

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: int
        :param reference: CapitainsCtsPassage Reference
        :type reference: CtsReference
        :param _debug: Check on passages duplicates
        :type _debug: bool
        :returns: List of levels

        .. note:: GetValidReff works for now as a loop using CapitainsCtsPassage, subinstances of CtsTextMetadata, to retrieve the valid \
        informations. Maybe something is more powerfull ?
        """

        depth = 0
        xml = self.textObject.xml

        if reference:
            if isinstance(reference, CtsReference):
                if not reference.is_range():
                    passages = [reference.start.list]
                    depth = len(passages[0])
                    if level == 0:
                        level = None
                        if _debug:
                            warnings.warn("Using level=0 with a Non-range Reference is invalid. Autocorrected to 1")
                else:
                    xml = self.getTextualNode(subreference=reference)

                    common = []

                    for index, part in enumerate(reference.start.list):
                        if index <= reference.end.depth:
                            if part == reference.end.list[index]:
                                common.append(part)
                            else:
                                break
                        else:
                            break

                    passages = [common]
                    depth = len(common)

                    if level is None:
                        level = reference.start.depth + depth
                    elif level == 1:
                        level = reference.start.depth + 1
                    elif level == 0:
                        level = reference.start.depth
            else:
                raise TypeError()
        else:
            passages = [[]]

        if level is None:
            level = 1

        if level <= len(passages[0]) and reference is not None:
            level = len(passages[0]) + 1
        if level > len(self.citation.root):
            raise CitationDepthError("The required level is too deep")

        nodes = [None] * (level - depth)

        citations = [citation for citation in self.citation.root]

        while len(nodes) >= 1:
            passages = [
                refs + [node.get(current_citation.attribute.replace("xml:", "{http://www.w3.org/XML/1998/namespace}"))]
                for xpath_result, refs, current_citation in [
                        (
                            xml.xpath(
                                citations[len(filling)-1].fill(filling),
                                namespaces=XPATH_NAMESPACES
                            ),
                            refs,
                            citations[len(filling)-1]
                        )
                        for filling, refs in
                        [(refs + [None], refs) for refs in passages]
                    ]
                for node in xpath_result
            ]
            nodes.pop(0)

            if len(passages) == 0:
                msg = "Unknown reference {}".format(reference)
                raise KeyError(msg)

        passages = [".".join(passage) for passage in passages]

        if _debug:
            duplicates = set()
            seen = set()
            for n in passages:
                if n in seen:
                    duplicates.add(n)
                else:
                    seen.add(n)
            if len(duplicates) > 0:
                message = ", ".join(duplicates)
                warnings.warn(message, DuplicateReference)
            del duplicates
            empties = [n for n in passages if n.rstrip('.') != n or n == '']
            if len(empties) > 0:
                message = '{} empty reference(s) at citation level {}'.format(len(empties), level)
                warnings.warn(message, EmptyReference)

        references = CtsReferenceSet(
            [CtsReference(reff) for reff in passages],
            citation=self.citation.root[level-1],
            level=level
        )
        return references

    def xpath(self, *args, **kwargs):
        """ Perform XPath on the passage XML

        :param args: Ordered arguments for etree._Element().xpath()
        :param kwargs: Named arguments
        :return: Result list
        :rtype: list(etree._Element)
        """
        if "smart_strings" not in kwargs:
            kwargs["smart_strings"] = False
        return self.resource.xpath(*args, **kwargs)

    def tostring(self, *args, **kwargs):
        """ Transform the CapitainsCtsPassage in XML string

        :param args: Ordered arguments for etree.tostring() (except the first one)
        :param kwargs: Named arguments
        :return:
        """
        return etree.tostring(self.resource, *args, **kwargs)


class _SimplePassage(_SharedMethods, PrototypeCtsPassage):
    """ CapitainsCtsPassage for simple and quick parsing of texts

    :param resource: Element representing the passage
    :type resource: etree._Element
    :param reference: CapitainsCtsPassage reference
    :type reference: CtsReference
    :param urn: URN of the source text or of the passage
    :type urn: URN
    :param citation: XmlCtsCitation scheme of the text
    :type citation: Citation
    :param text: CtsTextMetadata containing the passage
    :type text: CapitainsCtsText
    """
    def __init__(self, resource, reference, citation, text, urn=None):
        super(_SimplePassage, self).__init__(
            resource=resource,
            citation=citation,
            **_make_passage_kwargs(urn, reference)
        )
        self._text = text
        self._reference = reference
        self._children = None
        self._depth = 0
        if reference is not None:
            self._depth = reference.depth
        self._prev_next = None

    @property
    def reference(self):
        """ URN CapitainsCtsPassage CtsReference

        :return: CtsReference
        :rtype: CtsReference
        """
        return self._reference

    @reference.setter
    def reference(self, value):
        self._reference = value

    @property
    def childIds(self):
        """ Children of the passage

        :rtype: None, CtsReference
        :returns: Dictionary of chidren, where key are subreferences
        """
        if self.depth >= len(self.citation.root):
            return []
        elif self._children is not None:
            return self._children
        else:
            self._children = self.getReffs()
            return self._children

    def getReffs(self, level=1, subreference=None) -> CtsReferenceSet:
        """ Reference available at a given level

        :param level: Depth required. If not set, should retrieve first encountered level (1 based). 0 retrieves inside
            a range
        :param subreference: Subreference (optional)
        :returns: List of levels
        """
        level += self.depth
        if not subreference:
            subreference = self.reference
        return self.textObject.getValidReff(level, reference=subreference)

    def getTextualNode(self, subreference: CtsReference=None):
        """ Special GetPassage implementation for SimplePassage (Simple is True by default)

        :param subreference:
        :return:
        """
        if not isinstance(subreference, CtsReference):
            subreference = CtsReference(subreference)
        return self.textObject.getTextualNode(subreference)

    @property
    def nextId(self):
        """ Next passage

        :returns: Next passage at same level
        :rtype: None, CtsReference
        """
        return self.siblingsId[1]

    @property
    def prevId(self) -> Optional[CtsReference]:
        """ Get the Previous passage reference

        :returns: Previous passage reference at the same level
        :rtype: None, CtsReference
        """
        return self.siblingsId[0]

    @property
    def siblingsId(self) -> Tuple[CtsReference, CtsReference]:
        """ Siblings Identifiers of the passage

        :rtype: (str, str)
        """

        if not self._text:
            raise MissingAttribute("CapitainsCtsPassage was iniated without CtsTextMetadata object")
        if self._prev_next is not None:
            return self._prev_next

        document_references = self._text.getReffs(level=self.depth)

        range_length = 1
        if self.reference.is_range():
            range_length = len(self.getReffs())

        start = document_references.index(self.reference.start)

        if start == 0:
            # If the passage is already at the beginning
            _prev = None
        elif start - range_length < 0:
            _prev = document_references[0]
        else:
            _prev = document_references[start - 1]

        if start + 1 == len(document_references):
            # If the passage is already at the end
            _next = None
        elif start + range_length > len(document_references):
            _next = document_references[-1]
        else:
            _next = document_references[start + 1]

        self._prev_next = (_prev, _next)
        return self._prev_next

    @property
    def textObject(self):
        """ CtsTextMetadata Object. Required for NextPrev

        :rtype: CapitainsCtsText
        """
        return self._text


class CapitainsCtsText(_SharedMethods, PrototypeCtsText):
    """ Implementation of CTS tools for local files

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference._capitains_cts.URN
    :param resource: A resource
    :type resource: lxml.etree._Element
    :param citation: Highest XmlCtsCitation level
    :type citation: Citation
    :param autoreffs: Parse references on load (default : True)
    :type autoreffs: bool
    :ivar resource: lxml
    """

    def __init__(self, urn=None, citation=None, resource=None):
        super(CapitainsCtsText, self).__init__(urn=urn, citation=citation, resource=resource)

        if self.resource is not None:
            self._findCRefPattern(self.resource)

    def _findCRefPattern(self, xml):
        """ Find CRefPattern in the text and set object.citation
        :param xml: Xml Resource
        :type xml: lxml.etree._Element
        :return: None
        """
        if not self.citation.is_set():
            citation = xml.xpath("//tei:refsDecl[@n='CTS']", namespaces=XPATH_NAMESPACES)
            if len(citation):
                self.citation = Citation.ingest(resource=citation[0], xpath=".//tei:cRefPattern")
            else:
                raise MissingRefsDecl("No reference declaration (refsDecl) found.")

    def test(self):
        """ Parse the object and generate the children
        """
        try:
            xml = self.xml.xpath(self.citation.scope, namespaces=XPATH_NAMESPACES)
            if len(xml) == 0:
                msg = "Main citation scope does not result in any result ({0})".format(self.citation.scope)
                raise RefsDeclError(msg)
        except Exception as E:
            raise E


class CapitainsCtsPassage(_SharedMethods, PrototypeCtsPassage):
    """ CapitainsCtsPassage class for local texts which rebuilds the tree up to the passage.

        For design purposes, some people would prefer the output of GetPassage to be consistent. ContextPassage rebuilds
        the tree of the text up to the passage, keeping attributes of original nodes

        **Example** :  for a text with a citation scheme with following refsDecl :
        `/TEI/text/body/div[@type='edition']/div[@n='$1']/div[@n='$2']/l[@n='$3']` and a passage 1.1.1-1.2.3, this
        class will build an XML tree looking like the following

         .. code-block:: xml

            <TEI ...>
                <text ...>
                    <body ...>
                        <div type='edition' ...>
                            <div n='1' ...>

                                <div n='1' ...>
                                    <l n='1'>...</l>
                                    ...
                                </div>
                                <div n='2' ...>
                                    <l n='3'>...</l>
                                </div>
                            </div>
                        </div>
                    </body>
                </text>
            </TEI>

        :param reference: CapitainsCtsPassage reference
        :type reference: CtsReference
        :param urn: URN of the source text or of the passage
        :type urn: URN
        :param citation: XmlCtsCitation scheme of the text
        :type citation: XmlCtsCitation
        :param resource: Element representing the passage
        :type resource: etree._Element
        :param text: CtsTextMetadata containing the passage
        :type text: CtsTextMetadata

        .. note::
            .prev, .next, .first and .last won't run on passage with a range made of two different level, such as
            1.1-1.2.3 or 1-a.b. Those will raise `InvalidSiblingRequest`

    """
    def __init__(self, reference, urn=None, citation=None, resource=None, text=None):

        super(CapitainsCtsPassage, self).__init__(
            citation=citation,
            resource=resource,
            **_make_passage_kwargs(urn, reference)
        )
        if urn is not None and urn.reference is not None:
            reference = urn.reference
        self._reference = reference
        self._text = text
        self._children = None
        self._depth = self._depth_2 = 1

        if self.reference and self.reference.start:
            self._depth_2 = self._depth = self.reference.start.depth
        if self.reference.is_range() and self.reference.end:
            self._depth_2 = self.reference.end.depth

        self._prev_next = None  # For caching purpose

    @property
    def reference(self):
        """ CtsReference of the object
        """
        return self._reference

    @property
    def childIds(self):
        """ Children of the passage

        :rtype: None, CtsReference
        :returns: Dictionary of chidren, where key are subreferences
        """
        self._raise_depth()
        if self.depth >= len(self.citation.root):
            return []
        elif self._children is not None:
            return self._children
        else:
            self._children = self.getReffs()
            return self._children

    @property
    def nextId(self):
        """ Next passage

        :returns: Next passage at same level
        :rtype: None, CtsReference
        """
        return self.siblingsId[1]

    @property
    def prevId(self):
        """ Get the Previous passage reference

        :returns: Previous passage reference at the same level
        :rtype: None, CtsReference
        """
        return self.siblingsId[0]

    def _raise_depth(self):
        """ Simple check that raises an exception if the passage cannot run first, last, next or prev

        See object notes

        :raise: InvalidSiblingRequest
        """
        if self._depth != self._depth_2:
            raise InvalidSiblingRequest()

    @property
    def siblingsId(self) -> Tuple[CtsReference, CtsReference]:
        """ Siblings Identifiers of the passage

        :rtype: (str, str)
        """
        self._raise_depth()

        if not self._text:
            raise MissingAttribute("CapitainsCtsPassage was initiated without CtsTextMetadata object")
        if self._prev_next:
            return self._prev_next

        document_references = self._text.getReffs(level=self.depth)

        if self.reference.is_range():
            start, end = self.reference.start, self.reference.end
            range_length = len(self.getReffs(level=0))
        else:
            start = end = self.reference.start
            range_length = 1

        start = document_references.index(start)
        end = document_references.index(end)

        if start == 0:
            # If the passage is already at the beginning
            _prev = None
        elif start - range_length < 0:
            if start == end:
                _prev = document_references[0]
            else:
                _prev = "{}-{}".format(document_references[0], document_references[start-1])
        else:
            if start == end:
                _prev = document_references[start-1]
            else:
                _prev = "{}-{}".format(document_references[start-range_length], document_references[start-1])

        if start + 1 == len(document_references) or end + 1 == len(document_references):
            # If the passage is already at the end
            _next = None
        elif end + range_length >= len(document_references):
            if start == end:
                _next = document_references[-1]
            else:
                _next = "{}-{}".format(document_references[end+1], document_references[-1])
        else:
            if start == end:
                _next = document_references[end+1]
            else:
                _next = "{}-{}".format(document_references[end+1], document_references[end + range_length])

        self._prev_next = (CtsReference(_prev), CtsReference(_next))
        return self._prev_next

    @property
    def next(self):
        """ Next CapitainsCtsPassage (Interactive CapitainsCtsPassage)
        """
        if self.nextId is not None:
            return super(CapitainsCtsPassage, self).getTextualNode(subreference=self.nextId)

    @property
    def prev(self):
        """ Previous CapitainsCtsPassage (Interactive CapitainsCtsPassage)
        """
        if self.prevId is not None:
            return super(CapitainsCtsPassage, self).getTextualNode(subreference=self.prevId)

    def getTextualNode(self, subreference=None, *args, **kwargs):
        if not isinstance(subreference, CtsReference):
            subreference = CtsReference(subreference)
        X = super(CapitainsCtsPassage, self).getTextualNode(subreference=subreference)
        X._text = self._text
        return X

    @property
    def textObject(self):
        """ CtsTextMetadata Object. Required for NextPrev

        :rtype: CapitainsCtsText
        """
        return self._text



from typing import Union
from MyCapytain.common.reference import URN, BaseReferenceSet, BaseReference
from MyCapytain.common.constants import RDF_NAMESPACES
from ..metadata import Collection
from rdflib import Literal

from .inventory import CtsTextMetadata
from ..text import InteractiveTextualNode


__all__ = [
    "PrototypeCtsNode",
    "PrototypeCtsText",
    "PrototypeCtsPassage"
]


class PrototypeCtsNode(InteractiveTextualNode):
    """ Initiate a Resource object

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference._capitains_cts.URN
    :param metadata: Collection Information about the Item
    :type metadata: Collection
    :param citation: XmlCtsCitation system of the text
    :type citation: Citation
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int
    :param resource: Resource used to navigate through the textual graph

    :cvar default_exclude: Default exclude for exports
    """

    def __init__(self, urn: Union[URN, str] = None, **kwargs):
        super(PrototypeCtsNode, self).__init__(identifier=str(urn), **kwargs)
        self._urn = None

        if urn is not None:
            self.urn = urn

    @property
    def urn(self) -> URN:
        """ URN Identifier of the object

        """
        return self._urn

    @urn.setter
    def urn(self, value: Union[URN, str]):
        """ Set the urn

        :param value: URN to be saved
        :raises: *TypeError* when the value is not URN compatible

        """
        if isinstance(value, str):
            value = URN(value)
        elif not isinstance(value, URN):
            raise TypeError("New urn must be string or {} instead of {}".format(type(URN), type(value)))
        self._urn = value

    def get_cts_metadata(self, key: str, lang: str = None) -> Literal:
        """ Get easily a metadata from the CTS namespace

        :param key: CTS property to retrieve
        :param lang: Language in which it should be
        :return: Literal value of the CTS graph property
        """
        return self.metadata.get_single(RDF_NAMESPACES.CTS.term(key), lang)

    def getValidReff(self, level: int = 1, reference: BaseReference = None) -> BaseReferenceSet:
        """ Given a resource, CtsText will compute valid reffs

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :param reference: Subreference (optional)
        :returns: List of levels
        """
        raise NotImplementedError()

    def getLabel(self) -> Collection:
        """ Retrieve metadata about the text

        :rtype: Collection
        :returns: Retrieve Label informations in a Collection format
        """
        raise NotImplementedError()

    def set_metadata_from_collection(self, text_metadata: CtsTextMetadata):
        """ Set the object metadata using its collections recursively

        :param text_metadata: Object representing the current text as a collection
        :type text_metadata: CtsTextMetadata
        """
        edition, work, textgroup = tuple(([text_metadata] + text_metadata.parents)[:3])

        for node in textgroup.metadata.get(RDF_NAMESPACES.CTS.groupname):
            lang = node.language
            self.metadata.add(RDF_NAMESPACES.CTS.groupname, lang=lang, value=str(node))
            self.set_creator(str(node), lang)

        for node in work.metadata.get(RDF_NAMESPACES.CTS.title):
            lang = node.language
            self.metadata.add(RDF_NAMESPACES.CTS.title, lang=lang, value=str(node))
            self.set_title(str(node), lang)

        for node in edition.metadata.get(RDF_NAMESPACES.CTS.label):
            lang = node.language
            self.metadata.add(RDF_NAMESPACES.CTS.label, lang=lang, value=str(node))
            self.set_subject(str(node), lang)

        for node in edition.metadata.get(RDF_NAMESPACES.CTS.description):
            lang = node.language
            self.metadata.add(RDF_NAMESPACES.CTS.description, lang=lang, value=str(node))
            self.set_description(str(node), lang)

        if not self.citation.is_set() and edition.citation.is_set():
            self.citation = edition.citation


class PrototypeCtsPassage(PrototypeCtsNode):
    """ CapitainsCtsPassage objects possess metadata informations

    :param urn: A URN identifier
    :type urn: MyCapytain.common.reference._capitains_cts.URN
    :param metadata: Collection Information about the Item
    :type metadata: Collection
    :param citation: XmlCtsCitation system of the text
    :type citation: Citation
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int
    :param resource: Resource used to navigate through the textual graph

    :cvar default_exclude: Default exclude for exports
    """

    def __init__(self, **kwargs):
        super(PrototypeCtsPassage, self).__init__(**kwargs)

    @property
    def reference(self) -> BaseReference:
        return self.urn.reference


class PrototypeCtsText(PrototypeCtsNode):
    """ A CTS CtsText
    """

    def __init__(self, citation=None, metadata=None, **kwargs):
        super(PrototypeCtsText, self).__init__(citation=citation, metadata=metadata, **kwargs)
        self._cts = None

    @property
    def reffs(self) -> BaseReferenceSet:
        """ Get all valid reffs for every part of the CtsText

        :rtype: [str]
        """
        if not self._cts:
            self._cts = BaseReferenceSet(
                [reff for reffs in [self.getReffs(level=i) for i in range(1, len(self.citation) + 1)] for reff in reffs]
            )
        return self._cts



from ._base import BaseCitationSet, BaseCitation, BaseReference, BaseReferenceSet
from ..metadata import Metadata
from MyCapytain.common.constants import RDF_NAMESPACES


_dts = RDF_NAMESPACES.DTS
_cite_type_term = str(_dts.term("citeType"))
_cite_structure_term = str(_dts.term("citeStructure"))


class DtsReference(BaseReference):
    def __new__(cls, *refs, metadata: Metadata=None, type_: str=None):
        o = BaseReference.__new__(cls, *refs)
        if metadata:
            o._metadata = metadata
        else:
            o._metadata = Metadata()  # toDo : Figure how to deal with Refs ID in the Sparql Graph

        if type_:
            o.type = type_
        return o

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def type(self):
        return self.metadata.get_single(_dts.term("citeType"))

    @type.setter
    def type(self, value):
        self.metadata.set(_dts.term("citeType"), value)

    def __eq__(self, other):
        return super(DtsReference, self).__eq__(other) and \
            isinstance(other, DtsReference) and \
            self.type == other.type

    def __repr__(self):
        return "<DtsReference <{}> [{}]>".format(
            "><".join([str(x) for x in self if x]),
            self.type
        )


class DtsReferenceSet(BaseReferenceSet):
    def __contains__(self, item: str) -> bool:
        return BaseReferenceSet.__contains__(self, item) or \
               BaseReferenceSet.__contains__(self, DtsReference(item))

    def __eq__(self, other):
        return super(DtsReferenceSet, self).__eq__(other) and \
            self.level == other.level and \
            isinstance(other, DtsReferenceSet) and \
            self.citation == other.citation


class DtsCitation(BaseCitation):
    def __init__(self, name=None, children=None, root=None):
        super(DtsCitation, self).__init__(name=name, children=children, root=root)

    def __eq__(self, other):
        return isinstance(other, BaseCitation) and \
            self.name == other.name and \
            self.children == other.children and \
               (
                   (
                       not self.is_root() and
                       not other.is_root() and
                       self.root == other.root
                   )
                   or (
                       self.is_root() and other.is_root()
                   )
               )

    @classmethod
    def ingest(cls, resource, root=None, **kwargs):
        """ Ingest a dictionary of DTS Citation object (as parsed JSON-LD) and
        creates the Citation Graph

        :param resource: List of Citation objects from the
            DTS Collection Endpoint (as expanded JSON-LD)
        :type resource: dict
        :param root: Root of the citation tree
        :type root: BaseCitationSet
        :return:
        """
        cite = cls(
            name=resource.get(_cite_type_term, [{"@value": None}])[0]["@value"],  # Not really clean ?
            root=root
        )
        for subCite in resource.get(_cite_structure_term, []):
            cite.add_child(cls.ingest(subCite, root=root))
        return cite

    def match(self, passageId):
        raise NotImplementedError("Passage Match is not part of the DTS Standard Citation")


class DtsCitationSet(BaseCitationSet):
    """ Set of citations following the DTS model (Unlike CTS, one citation
    can have two or more children)

    """

    def __init__(self, children=None):
        super(DtsCitationSet, self).__init__(children=children)
        self._depth = None

    @property
    def depth(self) -> int:
        if self._depth:
            return self._depth
        return super(DtsCitationSet, self).depth

    @depth.setter
    def depth(self, value: int):
        self._depth = value

    def __repr__(self):
        return "<MyCapytain.common.reference.DtsCitationSet object at %s>" % id(self)

    CitationClass = DtsCitation

    def match(self, passageId):
        raise NotImplementedError("Passage Match is not part of the DTS Standard Citation")

    @classmethod
    def ingest(cls, resource):
        """ Ingest a list of DTS Citation object (as parsed JSON-LD) and
        creates the Citation Graph

        :param resource: List of Citation objects from the
            DTS Collection Endpoint (as expanded JSON-LD)
        :type resource: list
        :return: Citation Graph
        """
        _set = cls()
        for data in resource:
            _set.add_child(
                cls.CitationClass.ingest(data, root=_set)
            )
        return _set



"""

"""
import io
import logging
import os.path
from glob import glob
from math import ceil

from MyCapytain.common.reference._capitains_cts import CtsReference, URN
from MyCapytain.common.utils.xml import xmlparser
from MyCapytain.errors import InvalidURN, UnknownObjectError, UndispatchedTextError
from MyCapytain.resolvers.prototypes import Resolver
from MyCapytain.resolvers.utils import CollectionDispatcher
from MyCapytain.resources.collections.cts import XmlCtsTextInventoryMetadata, XmlCtsTextgroupMetadata, \
    XmlCtsWorkMetadata, XmlCtsCitation, XmlCtsTextMetadata as InventoryText, \
    XmlCtsTranslationMetadata, XmlCtsEditionMetadata, XmlCtsCommentaryMetadata
from MyCapytain.resources.prototypes.cts.inventory import CtsEditionMetadata, CtsTextgroupMetadata, CtsWorkMetadata, \
    CtsCommentaryMetadata, CtsTextInventoryCollection, CtsTranslationMetadata, CtsTextInventoryMetadata
from MyCapytain.resources.prototypes.cts.inventory import CtsTextInventoryCollection
from MyCapytain.resources.texts.local.capitains.cts import CapitainsCtsText


__all__ = [
    "CtsCapitainsLocalResolver"
]


class CtsCapitainsLocalResolver(Resolver):
    """ XML Folder Based resolver. CtsTextMetadata and metadata resolver based on local directories

    :param resource: Resource should be a list of folders retaining data as Capitains Guidelines Repositories
    :type resource: [str]
    :param name: Key used to differentiate Repository and thus enabling different repo to be used
    :type name: str
    :param logger: Logging object
    :type logger: logging

    :cvar TEXT_CLASS: CtsTextMetadata Class [not instantiated] to be used to parse Texts. Can be changed to support Cache for example
    :type TEXT_CLASS: class
    :cvar DEFAULT_PAGE: Default Page to show
    :cvar PER_PAGE: Tuple representing the minimal number of texts returned, the default number and the maximum number of texts returned


    """
    CLASSES = {
        "text": CapitainsCtsText,
        "edition": XmlCtsEditionMetadata,
        "translation": XmlCtsTranslationMetadata,
        "commentary": XmlCtsCommentaryMetadata,
        "work": XmlCtsWorkMetadata,
        "textgroup": XmlCtsTextgroupMetadata,
        "inventory": XmlCtsTextInventoryMetadata,
        "inventory_collection": CtsTextInventoryCollection,
        "citation": XmlCtsCitation
    }

    DEFAULT_PAGE = 1
    PER_PAGE = (1, 10, 100)  # Min, Default, Mainvex,
    RAISE_ON_UNDISPATCHED = False
    RAISE_ON_GENERIC_PARSING_ERROR = True

    @property
    def inventory(self):
        return self.__inventory__

    @inventory.setter
    def inventory(self, value):
        self.__inventory__ = value

    @property
    def texts(self):
        return self.inventory.readableDescendants

    def __init__(self, resource, name=None, logger=None, dispatcher=None, autoparse=True):
        """ Initiate the XMLResolver
        """
        self.classes = {}
        self.classes.update(type(self).CLASSES)

        if dispatcher is None:
            inventory_collection = self.classes["inventory_collection"](identifier="defaultTic")
            ti = self.classes["inventory"]("default")
            ti.parent = inventory_collection
            ti.set_label("Default collection", "eng")
            self.dispatcher = CollectionDispatcher(inventory_collection)
        else:
            self.dispatcher = dispatcher
        self.__inventory__ = self.dispatcher.collection
        self.name = name

        self.logger = logger
        if not logger:
            self.logger = logging.getLogger(name)

        if not name:
            self.name = "repository"

        self.works = []

        if autoparse:
            self.parse(resource)

    def xmlparse(self, file):
        """ Parse a XML file
        :param file: Opened File
        :return: Tree
        """
        return xmlparser(file)

    def read(self, identifier, path):
        """ Retrieve and parse a text given an identifier

        :param identifier: Identifier of the text
        :type identifier: str
        :param path: Path of the text
        :type path: str
        :return: Parsed Text
        :rtype: CapitainsCtsText
        """
        with open(path) as f:
            o = self.classes["text"](urn=identifier, resource=self.xmlparse(f))
        return o

    def _parse_textgroup_wrapper(self, cts_file):
        """ Wraps with a Try/Except the textgroup parsing from a cts file

        :param cts_file: Path to the CTS File
        :type cts_file: str
        :return: CtsTextgroupMetadata
        """
        try:
            return self._parse_textgroup(cts_file)
        except Exception as E:
            self.logger.error("Error parsing %s ", cts_file)
            if self.RAISE_ON_GENERIC_PARSING_ERROR:
                raise E

    def _parse_textgroup(self, cts_file):
        """ Parses a textgroup from a cts file

        :param cts_file: Path to the CTS File
        :type cts_file: str
        :return: CtsTextgroupMetadata and Current file
        """
        with io.open(cts_file) as __xml__:
            return self.classes["textgroup"].parse(
                resource=__xml__
            ), cts_file

    def _parse_work_wrapper(self, cts_file, textgroup):
        """ Wraps with a Try/Except the Work parsing from a cts file

        :param cts_file: Path to the CTS File
        :type cts_file: str
        :param textgroup: Textgroup to which the Work is a part of
        :type textgroup: CtsTextgroupMetadata
        :return: Parsed Work and the Texts, as well as the current file directory
        """
        try:
            return self._parse_work(cts_file, textgroup)
        except Exception as E:
            self.logger.error("Error parsing %s ", cts_file)
            if self.RAISE_ON_GENERIC_PARSING_ERROR:
                raise E

    def _parse_work(self, cts_file, textgroup):
        """ Parses a work from a cts file

        :param cts_file: Path to the CTS File
        :type cts_file: str
        :param textgroup: Textgroup to which the Work is a part of
        :type textgroup: CtsTextgroupMetadata
        :return: Parsed Work and the Texts, as well as the current file directory
        """
        with io.open(cts_file) as __xml__:
            work, texts = self.classes["work"].parse(
                resource=__xml__,
                parent=textgroup,
                _with_children=True
            )

        return work, texts, os.path.dirname(cts_file)

    def _parse_text(self, text, directory):
        """ Complete the TextMetadata object with its citation scheme by parsing the original text

        :param text: Text Metadata collection
        :type text: XmlCtsTextMetadata
        :param directory: Directory in which the metadata was found and where the text file should be
        :type directory: str
        :returns: True if all went well
        :rtype: bool
        """
        text_id, text_metadata = text.id, text
        text_metadata.path = "{directory}/{textgroup}.{work}.{version}.xml".format(
            directory=directory,
            textgroup=text_metadata.urn.textgroup,
            work=text_metadata.urn.work,
            version=text_metadata.urn.version
        )
        if os.path.isfile(text_metadata.path):
            try:
                text = self.read(text_id, path=text_metadata.path)
                cites = list()
                for cite in [c for c in text.citation][::-1]:
                    if len(cites) >= 1:
                        cites.append(self.classes["citation"](
                            xpath=cite.xpath.replace("'", '"'),
                            scope=cite.scope.replace("'", '"'),
                            name=cite.name,
                            child=cites[-1]
                        ))
                    else:
                        cites.append(self.classes["citation"](
                            xpath=cite.xpath.replace("'", '"'),
                            scope=cite.scope.replace("'", '"'),
                            name=cite.name
                        ))
                del text
                text_metadata.citation = cites[-1]
                self.logger.info("%s has been parsed ", text_metadata.path)
                if not text_metadata.citation.is_set():
                    self.logger.error("%s has no passages", text_metadata.path)
                    return False
                return True
            except Exception:
                self.logger.error(
                    "%s does not accept parsing at some level (most probably citation) ",
                    text_metadata.path
                )
                return False
        else:
            self.logger.error("%s is not present", text_metadata.path)
            return False

    def _dispatch(self, textgroup, directory):
        """ Run the dispatcher over a textgroup.

        :param textgroup: Textgroup object that needs to be dispatched
        :param directory: Directory in which the textgroup was found
        """
        if textgroup.id in self.dispatcher.collection:
            self.dispatcher.collection[textgroup.id].update(textgroup)
        else:
            self.dispatcher.dispatch(textgroup, path=directory)

        for work_urn, work in textgroup.works.items():
            if work_urn in self.dispatcher.collection[textgroup.id].works:
                self.dispatcher.collection[work_urn].update(work)

    def _dispatch_container(self, textgroup, directory):
        """ Run the dispatcher over a textgroup within a try/except block

        .. note:: This extraction allows to change the dispatch routine \
            without having to care for the error dispatching

        :param textgroup: Textgroup object that needs to be dispatched
        :param directory: Directory in which the textgroup was found
        """
        try:
            self._dispatch(textgroup, directory)
        except UndispatchedTextError as E:
            self.logger.error("Error dispatching %s ", directory)
            if self.RAISE_ON_UNDISPATCHED is True:
                raise E

    def _clean_invalids(self, invalids):
        """ Optionally remove texts that were found to be invalid

        :param invalids: List of text identifiers
        :type invalids: [CtsTextMetadata]
        """
        pass

    def parse(self, resource):
        """ Parse a list of directories and reads it into a collection

        :param resource: List of folders
        :return: An inventory resource and a list of CtsTextMetadata metadata-objects
        """
        textgroups = []
        texts = []
        invalids = []

        for folder in resource:
            cts_files = glob("{base_folder}/data/*/__cts__.xml".format(base_folder=folder))
            for cts_file in cts_files:
                textgroup, cts_file = self._parse_textgroup_wrapper(cts_file)
                textgroups.append((textgroup, cts_file))

        for textgroup, cts_textgroup_file in textgroups:
            cts_work_files = glob("{parent}/*/__cts__.xml".format(parent=os.path.dirname(cts_textgroup_file)))

            for cts_work_file in cts_work_files:
                _, parsed_texts, directory = self._parse_work_wrapper(cts_work_file, textgroup)
                texts.extend([(text, directory) for text in parsed_texts])

        for text, directory in texts:
            # If text_id is not none, the text parsing errored
            if not self._parse_text(text, directory):
                invalids.append(text)

        # Dispatching routine
        for textgroup, textgroup_path in textgroups:
            self._dispatch_container(textgroup, textgroup_path)

        # Clean invalids if there was a need
        self._clean_invalids(invalids)

        self.inventory = self.dispatcher.collection
        return self.inventory

    def __getText__(self, urn):
        """ Returns a CtsTextMetadata object
        :param urn: URN of a text to retrieve
        :type urn: str, URN
        :return: Textual resource and metadata
        :rtype: (CapitainsCtsText, InventoryText)
        """
        if not isinstance(urn, URN):
            urn = URN(urn)
        if len(urn) != 5:
            if len(urn) == 4:
                urn, reference = urn.upTo(URN.WORK), str(urn.reference)
                urn = [
                    t.id
                    for t in self.texts
                    if t.id.startswith(str(urn)) and isinstance(t, CtsEditionMetadata)
                ]
                if len(urn) > 0:
                    urn = URN(urn[0])
                else:
                    raise UnknownObjectError
            else:
                raise InvalidURN

        text = self.inventory[str(urn)]

        if os.path.isfile(text.path):
            with io.open(text.path) as __xml__:
                resource = self.classes["text"](urn=urn, resource=self.xmlparse(__xml__))
        else:
            resource = None
            self.logger.warning('The file {} is mentioned in the metadata but does not exist'.format(text.path))

        return resource, text

    def __getTextMetadata__(self,
                            urn=None, page=None, limit=None,
                            lang=None, category=None, pagination=False
                            ):
        """ Retrieve a slice of the inventory filtered by given arguments
        :param urn: Partial URN to use to filter out resources
        :type urn: str
        :param page: Page to show
        :type page: int
        :param limit: Item Per Page
        :type limit: int
        :param inventory: Inventory name
        :type inventory: str
        :param lang: Language to filter on
        :type lang: str
        :param category: Type of elements to show
        :type category: str
        :param pagination: Activate pagination
        :type pagination: bool
        :return: ([Matches], Page, Count)
        :rtype: ([CtsTextMetadata], int, int)
        """
        __PART = None
        if urn is not None:
            if isinstance(urn, URN):
                _urn = urn
            else:
                _urn = URN(urn)
            __PART = [None, None, URN.NAMESPACE, URN.TEXTGROUP, URN.WORK, URN.VERSION, URN.COMPLETE][len(_urn)]

        matches = [
            text
            for text in self.texts
            if
            (lang is None or (lang is not None and lang == text.lang)) and
            (urn is None or (urn is not None and text.urn.upTo(__PART) == urn)) and
            (text.citation is not None) and
            (
                category not in ["edition", "translation", "commentary"] or
                (category in ["edition", "translation", "commentary"] and category.lower() == text.subtype.lower())
            )
        ]
        if pagination:
            start_index, end_index, page, count = type(self).pagination(page, limit, len(matches))
        else:
            start_index, end_index, page, count = None, None, 0, len(matches)

        return matches[start_index:end_index], page, count

    @staticmethod
    def pagination(page, limit, length):
        """ Help for pagination
        :param page: Provided Page
        :param limit: Number of item to show
        :param length: Length of the list to paginate
        :return: (Start Index, End Index, Page Number, Item Count)
        """
        realpage = page
        page = page or CtsCapitainsLocalResolver.DEFAULT_PAGE
        limit = limit or CtsCapitainsLocalResolver.PER_PAGE[1]

        if limit < CtsCapitainsLocalResolver.PER_PAGE[0] or limit > CtsCapitainsLocalResolver.PER_PAGE[2]:
            limit = CtsCapitainsLocalResolver.PER_PAGE[1]

        page = (page - 1) * limit

        if page > length:
            realpage = int(ceil(length / limit))
            page = limit * (realpage - 1)
            count = length - 1
        elif limit - 1 + page < length:
            count = limit - 1 + page
        else:
            count = length - 1

        return page, count + 1, realpage, count - page + 1

    def getMetadata(self, objectId=None, **filters):
        """ Request metadata about a text or a collection

        :param objectId: Object Identifier to filter on
        :type objectId: str
        :param filters: Kwargs parameters.
        :type filters: dict
        :return: Collection
        """
        if objectId is None:
            return self.inventory
        elif objectId in self.inventory.children.keys():
            return self.inventory[objectId]
        texts, _, _ = self.__getTextMetadata__(urn=objectId)

        # We store inventory names and if there is only one we recreate the inventory
        inv_names = [text.parent.parent.parent.id for text in texts]
        if len(set(inv_names)) == 1:
            inventory = self.classes["inventory"](name=inv_names[0])
        else:
            inventory = self.classes["inventory"]()

        # For each text we found using the filter
        for text in texts:
            tg_urn = str(text.parent.parent.urn)
            wk_urn = str(text.parent.urn)
            txt_urn = str(text.urn)
            # If we need to generate a textgroup object
            if tg_urn not in inventory.textgroups:
                self.classes["textgroup"](urn=tg_urn, parent=inventory)
            # If we need to generate a work object
            if wk_urn not in inventory.textgroups[tg_urn].works:
                self.classes["work"](urn=wk_urn, parent=inventory.textgroups[tg_urn])

            if isinstance(text, CtsEditionMetadata):
                x = self.classes["edition"](urn=txt_urn, parent=inventory.textgroups[tg_urn].works[wk_urn])
                x.citation = text.citation
            elif isinstance(text, CtsTranslationMetadata):
                x = self.classes["translation"](urn=txt_urn, parent=inventory.textgroups[tg_urn].works[wk_urn], lang=text.lang)
                x.citation = text.citation
            elif isinstance(text, CtsCommentaryMetadata):
                x = self.classes["commentary"](urn=txt_urn, parent=inventory.textgroups[tg_urn].works[wk_urn], lang=text.lang)
                x.citation = text.citation

        return inventory[objectId]

    def getTextualNode(self, textId, subreference=None, prevnext=False, metadata=False):
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage CtsReference
        :type subreference: str
        :param prevnext: Retrieve graph representing previous and next passage
        :type prevnext: boolean
        :param metadata: Retrieve metadata about the passage and the text
        :type metadata: boolean
        :return: CapitainsCtsPassage
        :rtype: CapitainsCtsPassage
        """
        text, text_metadata = self.__getText__(textId)
        if subreference is not None and not isinstance(subreference, CtsReference):
            subreference = CtsReference(subreference)
        passage = text.getTextualNode(subreference)
        if metadata:
            passage.set_metadata_from_collection(text_metadata)
        return passage

    def getSiblings(self, textId, subreference: CtsReference):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage CtsReference
        :type subreference: str
        :return: Tuple of references
        :rtype: (str, str)
        """
        text, inventory = self.__getText__(textId)
        if not isinstance(subreference, CtsReference):
            subreference = CtsReference(subreference)
        passage = text.getTextualNode(subreference)
        return passage.siblingsId

    def getReffs(self, textId, level=1, subreference=None):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage CtsReference
        :type subreference: str
        :return: List of references
        :rtype: [str]
        """
        passage, inventory = self.__getText__(textId)
        if subreference:
            passage = passage.getTextualNode(subreference)
        return passage.getReffs(level=level, subreference=subreference)



import re
from pyld.jsonld import expand
from typing import Optional

from MyCapytain.common.constants import RDF_NAMESPACES
from MyCapytain.errors import UnknownCollection
from ._base import DtsCollection


_hyd = RDF_NAMESPACES.HYDRA
_empty = [{"@value": None}]
_re_page = re.compile("page=(\d+)")


class PaginatedProxy:

    __UPDATES_CALLABLES__ = {
        "update", "add", "extend"
    }

    def __init__(
            self,
            obj,
            proxied: str,
            update_lambda,
            condition_lambda
    ):
        self._proxied = getattr(obj, proxied)
        self._obj = obj
        self._attr = proxied
        self._condition_lambda = condition_lambda
        self._update_lambda = update_lambda

    def __getattr__(self, item):
        if item in self.__UPDATES_CALLABLES__:
            return getattr(self._proxied, item)
        if item == "set":
            return self.set
        return self._run(item)

    def _run(self, item: Optional[str]=None):
        if not self._condition_lambda():
            self._update_lambda()

        if item:
            return getattr(self._proxied, item)
        return self._proxied

    def set(self, value) -> None:
        self._proxied = value

    def __iter__(self):
        return iter(self._run())

    def __getitem__(self, item):
        if isinstance(self._proxied, (list, dict, set, tuple)):
            # If it is in, we do not update the object
            if item in self._proxied:
                return self._proxied[item]
            # If it not in, and we are still in this object,
            # It means we need to crawl :
            self._run()
            return self._proxied[item]
        raise TypeError("'PaginatedProxy' object is not subscriptable")

    def __eq__(self, other):
        return self._proxied == other


class HttpResolverDtsCollection(DtsCollection):
    def __init__(
            self,
            identifier: str,
            resolver: "HttpDtsResolver",
            metadata_parsed=True, *args, **kwargs):
        super(HttpResolverDtsCollection, self).__init__(identifier, *args, **kwargs)

        self._parsed = {
            "children": False,
            "parents": False,
            "metadata": metadata_parsed
        }
        self._last_page_parsed = {
            "children": None,
            "parents": None,
        }

        self._children = PaginatedProxy(
            self,
            "_children",
            update_lambda=lambda: self._parse_paginated_members(direction="children"),
            condition_lambda=lambda: self._parsed["parents"]
        )
        self._parents = PaginatedProxy(
            self,
            "_parents",
            update_lambda=lambda: self._parse_paginated_members(direction="parents"),
            condition_lambda=lambda: self._parsed["parents"]
        )

        self._resolver = resolver

    def _parse_paginated_members(self, direction="children"):
        """ Launch parsing of children
        """

        page = self._last_page_parsed[direction]
        if not page:
            page = 1
        else:
            page = int(page)
        while page:
            if page > 1:
                response = self._resolver.endpoint.get_collection(
                    collection_id=self.id,
                    page=page,
                    nav=direction
                )
            else:
                response = self._resolver.endpoint.get_collection(
                    collection_id=self.id,
                    nav=direction
                )
            response.raise_for_status()

            data = response.json()
            data = expand(data)[0]

            if direction == "children":
                self.children.update({
                    o.id: o
                    for o in type(self).parse_member(
                        obj=data, collection=self, direction=direction, resolver=self._resolver
                    )
                })
            else:
                self.parents.update({
                    o
                    for o in type(self).parse_member(
                        obj=data, collection=self, direction=direction, resolver=self._resolver
                    )
                })
            self._last_page_parsed[direction] = page

            page = None
            if "https://www.w3.org/ns/hydra/core#view" in data:
                if "https://www.w3.org/ns/hydra/core#next" in data["https://www.w3.org/ns/hydra/core#view"][0]:
                    page = int(_re_page.findall(
                        data["https://www.w3.org/ns/hydra/core#view"]
                            [0]["https://www.w3.org/ns/hydra/core#next"]
                            [0]["@value"]
                    )[0])

        self._parsed[direction] = True

    @property
    def children(self):
        if not self._parsed["children"]:
            if not self.size == 0:
                return self._children
            else:
                self._parsed["children"] = True
        if isinstance(self._children, PaginatedProxy):

            self._children = self._children._proxied

        return self._children

    @property
    def parents(self):
        if self._parsed["parents"] and isinstance(self._parents, PaginatedProxy):
            self._parents = self._parents._proxied
        return self._parents

    def retrieve(self):
        if not self._parsed["metadata"]:
            query = self._resolver.endpoint.get_collection(self.id)
            data = query.json()
            if not len(data):
                raise UnknownCollection(
                    "The contacted endpoint seems to not have any data about collection %s " % self.id
                )
            self._parse_metadata(expand(data)[0])
        return True

    @classmethod
    def parse_member(
            cls,
            obj: dict,
            collection: "HttpResolverDtsCollection",
            direction: str,
            **additional_parameters):

        """ Parse the member value of a Collection response
        and returns the list of object while setting the graph
        relationship based on `direction`

        :param obj: PyLD parsed JSON+LD
        :param collection: Collection attached to the member property
        :param direction: Direction of the member (children, parent)
        """
        members = []

        # Start pagination check here
        hydra_members = obj.get(str(_hyd.member), [])

        if hydra_members:
            for member in hydra_members:
                subcollection = cls.parse(member, metadata_parsed=False, **additional_parameters)
                if direction == "children":
                    subcollection.parents.set({collection})
                members.append(subcollection)

            if "https://www.w3.org/ns/hydra/core#view" not in obj or \
                    (direction == "children" and collection.size == 0):
                collection._parsed[direction] = True

        return members



import re
from collections import OrderedDict
from functools import reduce


class OrderedDefaultDict(OrderedDict):
    """ Extension of Default Dict that makes an OrderedDefaultDict

    :param default_factory__: Default class to initiate
    """

    def __init__(self, default_factory=None, *args, **kwargs):
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        val = self[key] = self.default_factory()
        return val


def nested_ordered_dictionary():
    """ Helper to create a nested ordered default dictionary

    :rtype OrderedDefaultDict:
    :return: Nested Ordered Default Dictionary instance
    """
    return OrderedDefaultDict(nested_ordered_dictionary)


def nested_get(dictionary, keys):
    """ Get value in dictionary for dictionary[keys[0]][keys[1]][keys[..n]]

    :param dictionary: An input dictionary
    :param keys: Keys where to store data
    :return:
    """
    return reduce(lambda d, k: d[k], keys, dictionary)


def nested_set(dictionary,  keys, value):
    """ Set value in dictionary for dictionary[keys[0]][keys[1]][keys[..n]]

    :param dictionary: An input dictionary
    :param keys: Keys where to store data
    :param value: Value to set at keys** target
    :return: None
    """
    nested_get(dictionary, keys[:-1])[keys[-1]] = value


_strip = re.compile("([ ]{2,})+")


def normalize(string):
    """ Remove double-or-more spaces in a string

    :param string: A string to change
    :type string: text_type
    :rtype: text_type
    :returns: Clean string
    """
    return _strip.sub(" ", string)


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.proto.text
   :synopsis: Prototypes for CtsText

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from typing import Union, List, Iterator
from rdflib.namespace import DC, DCTERMS as DCT
from rdflib import BNode, URIRef, Graph, Literal
from rdflib.term import Identifier
from MyCapytain.common.reference import Citation, NodeId, BaseReference, BaseReferenceSet
from MyCapytain.common.metadata import Metadata
from MyCapytain.common.constants import Mimetypes, get_graph, RDF_NAMESPACES
from MyCapytain.common.base import Exportable
from MyCapytain.resources.prototypes.metadata import Collection


__all__ = [
    "TextualElement",
    "TextualGraph",
    "TextualNode",
    "InteractiveTextualNode"
]


class TextualElement(Exportable):
    """ Node representing a text passage.

    :param identifier: Identifier of the text
    :type identifier: str
    :param metadata: Collection Information about the Item
    :type metadata: Collection

    :cvar default_exclude: Default exclude for exports
    """

    default_exclude = []

    def __init__(self, identifier: str=None, metadata: Metadata=None):
        self._graph = get_graph()
        self._identifier = identifier

        self._node = BNode()
        self._metadata = metadata or Metadata(node=self.asNode())

        self._graph.addN([
            (self._node, RDF_NAMESPACES.DTS.implements, URIRef(identifier), self._graph)#,
            #(self.__node__, RDF_NAMESPACES.DTS.metadata, self.metadata.asNode(), self.__graph__)
        ])

    def __repr__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, self.id)

    @property
    def graph(self) -> Graph:
        return self._graph

    @property
    def text(self) -> str:
        """ String representation of the text

        :return: String representation of the text
        """
        return self.export(output=Mimetypes.PLAINTEXT, exclude=self.default_exclude)

    @property
    def id(self) -> str:
        """ Identifier of the text

        :return: Identifier of the text
        """
        return self._identifier

    @property
    def metadata(self) -> Metadata:
        """ Metadata information about the text

        :return: Collection object with metadata about the text
        :rtype: Metadata
        """
        return self._metadata

    def asNode(self) -> Identifier:
        return self._node

    def get_creator(self, lang: str=None) -> Literal:
        """ Get the DC Creator literal value

        :param lang: Language to retrieve
        :return: Creator string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.creator, lang=lang)

    def set_creator(self, value: Union[Literal, Identifier, str], lang: str= None):
        """ Set the DC Creator literal value

        :param value: Value of the creator node
        :param lang: Language in which the value is
        """
        self.metadata.add(key=DC.creator, value=value, lang=lang)

    def get_title(self, lang: str=None) -> Literal:
        """ Get the title of the object

        :param lang: Lang to retrieve
        :return: Title string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.title, lang=lang) or \
            self.metadata.get_single(key=DCT.title, lang=lang)

    def set_title(self, value: Union[Literal, Identifier, str], lang: str= None):
        """ Set the DC Title literal value

        :param value: Value of the title node
        :param lang: Language in which the value is
        """
        return self.metadata.add(key=DC.title, value=value, lang=lang)

    def get_description(self, lang: str=None) -> Literal:
        """ Get the description of the object

        :param lang: Lang to retrieve
        :return: Description string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.description, lang=lang)

    def set_description(self, value: Union[Literal, Identifier, str], lang: str= None):
        """ Set the DC Description literal value

        :param value: Value of the title node
        :param lang: Language in which the value is
        """
        return self.metadata.add(key=DC.description, value=value, lang=lang)

    def get_subject(self, lang=None) -> Literal:
        """ Get the subject of the object

        :param lang: Lang to retrieve
        :return: Subject string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.subject, lang=lang)

    def set_subject(self, value: Union[Literal, Identifier, str], lang: str= None):
        """ Set the DC Subject literal value

        :param value: Value of the subject node
        :param lang: Language in which the value is
        """
        return self.metadata.add(key=DC.subject, value=value, lang=lang)

    def export(self, output: str=None, exclude: List[str]=None, **kwargs):
        """ Export the collection item in the Mimetype required.

        ..note:: If current implementation does not have special mimetypes, reuses default_export method

        :param output: Mimetype to export to (Uses MyCapytain.common.constants.Mimetypes)
        :type output: str
        :param exclude: Information to exclude. Specific to implementations
        :type exclude: [str]
        :return: Object using a different representation
        """
        return Exportable.export(self, output, exclude=exclude, **kwargs)


class TextualNode(TextualElement, NodeId):
    """ Node representing a text passage.

    :param identifier: Identifier of the text
    :type identifier: str
    :param metadata: Collection Information about the Item
    :type metadata: Collection
    :param citation: XmlCtsCitation system of the text
    :type citation: Citation
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int

    :cvar default_exclude: Default exclude for exports
    """

    def __init__(self, identifier: str=None, citation: Citation=None, **kwargs):
        super(TextualNode, self).__init__(identifier=identifier, **kwargs)
        self._citation = citation or Citation()

    @property
    def citation(self) -> Citation:
        """ Citation system of the object

        :rtype: Citation
        """
        return self._citation

    @citation.setter
    def citation(self, value: Citation):
        if not isinstance(value, Citation):
            raise TypeError("Citation property can only host a Citation object")
        self._citation = value


class TextualGraph(TextualNode):
    """ Node representing a text passage.

    :param identifier: Identifier of the text
    :type identifier: str
    :param metadata: Collection Information about the Item
    :type metadata: Collection
    :param citation: XmlCtsCitation system of the text
    :type citation: Citation
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int
    :param resource: Resource used to navigate through the textual graph

    :cvar default_exclude: Default exclude for exports
    """
    def __init__(self, identifier: str=None, **kwargs):
        super(TextualGraph, self).__init__(identifier=identifier, **kwargs)

    def getTextualNode(self, subreference: BaseReference) -> "TextualGraph":
        """ Retrieve a passage and store it in the object

        :param subreference: CtsReference of the passage to retrieve
        :returns: Object representing the passage

        :raises: *TypeError* when reference is not a list or a CtsReference
        """

        raise NotImplementedError()

    def getReffs(self, level: int=1, subreference: BaseReference=None) -> BaseReferenceSet:
        """ CtsReference available at a given level

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :param subreference: Subreference (optional)
        :returns: List of levels
        """
        raise NotImplementedError()


class InteractiveTextualNode(TextualGraph):
    """ Node representing a text passage.

    :param identifier: Identifier of the text
    :type identifier: str
    :param metadata: Collection Information about the Item
    :type metadata: Collection
    :param citation: XmlCtsCitation system of the text
    :type citation: Citation
    :param children: Current node Children's Identifier
    :type children: [str]
    :param parent: Parent of the current node
    :type parent: str
    :param siblings: Previous and next node of the current node
    :type siblings: str
    :param depth: Depth of the node in the global hierarchy of the text tree
    :type depth: int
    :param resource: Resource used to navigate through the textual graph

    :cvar default_exclude: Default exclude for exports
    """
    def __init__(self, identifier: str=None, **kwargs):
        super(InteractiveTextualNode, self).__init__(identifier=identifier, **kwargs)
        self._childIds = None

    @property
    def prev(self) -> "InteractiveTextualNode":
        """ Get Previous TextualNode
        """
        if self.prevId is not None:
            return self.getTextualNode(self.prevId)

    @property
    def next(self) -> "InteractiveTextualNode":
        """ Get Next TextualNode
        """
        if self.nextId is not None:
            return self.getTextualNode(self.nextId)

    @property
    def children(self) -> Iterator["InteractiveTextualNode"]:
        """ Children TextualNode

        """
        for ID in self.childIds:
            yield self.getTextualNode(ID)

    @property
    def parent(self) -> "InteractiveTextualNode":
        """ Parent TextualNode

        """
        return self.getTextualNode(self.parentId)

    @property
    def first(self) -> TextualNode:
        """ First TextualNode
        """
        if self.firstId is not None:
            return self.getTextualNode(self.firstId)

    @property
    def last(self) -> TextualNode:
        """ Last CapitainsCtsPassage
        """
        if self.lastId is not None:
            return self.getTextualNode(self.lastId)

    @property
    def childIds(self) -> BaseReferenceSet:
        """ Identifiers of children

        :return: Identifiers of children
        """
        if self._childIds is None:
            self._childIds = self.getReffs()
        return self._childIds

    @property
    def firstId(self) -> BaseReference:
        """ First child's id of current TextualNode
        """
        if self.childIds is not None:
            if len(self.childIds) > 0:
                return self.childIds[0]
            return None
        else:
            raise NotImplementedError

    @property
    def lastId(self) -> BaseReference:
        """ Last child's id of current TextualNode
        """
        if self.childIds is not None:
            if len(self.childIds) > 0:
                return self.childIds[-1]
            return None
        else:
            raise NotImplementedError



from collections import namedtuple
from urllib.parse import parse_qs, urlparse, urljoin

import link_header

_Route = namedtuple("Route", ["path", "query_dict"])
_Navigation = namedtuple("Navigation", ["prev", "next", "last", "current", "first"])


def parse_pagination(headers):
    """ Parses headers to create a pagination objects

    :param headers: HTTP Headers
    :type headers: dict
    :return: Navigation object for pagination
    :rtype: _Navigation
    """
    links = {
        link.rel: parse_qs(link.href).get("page", None)
        for link in link_header.parse(headers.get("Link", "")).links
    }
    return _Navigation(
        links.get("previous", [None])[0],
        links.get("next", [None])[0],
        links.get("last", [None])[0],
        links.get("current", [None])[0],
        links.get("first", [None])[0]
    )


def parse_uri(uri, endpoint_uri):
    """ Parse a URI into a Route namedtuple

    :param uri: URI or relative URI
    :type uri: str
    :param endpoint_uri: URI of the endpoint
    :type endpoint_uri: str
    :return: Parsed route
    :rtype: _Route
    """
    temp_parse = urlparse(uri)
    return _Route(
        urljoin(endpoint_uri, temp_parse.path),
        parse_qs(temp_parse.query)
    )



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain
   :platform: Unix, Windows
   :synopsis: Canonical CtsTextMetadata Service and Canonical URN tool suite for Python

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

__version__ = "3.0.1"



import re
from typing import Optional, List, Union, Tuple
from lxml.etree import _Element

from MyCapytain.common.constants import Mimetypes, get_graph, RDF_NAMESPACES, XPATH_NAMESPACES
from MyCapytain.common.utils.xml import make_xml_node

from ._base import BaseCitation, BaseReference, BaseReferenceSet

REFSDECL_SPLITTER = re.compile(r"/+[*()|\sa-zA-Z0-9:\[\]@=\\{$'\".\s]+")
REFSDECL_REPLACER = re.compile(r"\$[0-9]+")
SUBREFERENCE = re.compile(r"(\w*)\[?([0-9]*)\]?", re.UNICODE)
REFERENCE_REPLACER = re.compile(r"(@[a-zA-Z0-9:]+)(=)([\\$'\"?0-9]{3,6})")


def _child_or_none(liste):
    """ Used to parse resources in XmlCtsCitation

    :param liste: List of item
    :return: If there is > 1 element in the list, return the last one
    """
    if len(liste) > 0:
        return liste[-1]
    else:
        return None


class CtsWordReference(str):
    """ A CTSWordReference is the specific part of a CTS identifier that
    identifies a word in a given passage. It contains the text in its
    .word property and the index of this word in its .counter identifier

    It can be returned as a tuple using .tuple()

    """
    def __new__(cls, word_reference: str):
        word, counter = tuple(SUBREFERENCE.findall(word_reference)[0])

        if counter:
            counter = int(counter)
        else:
            counter = 0

        obj = str.__new__(cls, "@"+word_reference)
        obj.counter = counter
        obj.word = word

        return obj

    def tuple(self) -> Tuple[str, int]:
        return self.word, self.counter

    def __iter__(self):
        return iter([self.word, self.counter])


class CtsSinglePassageId(str):
    """ A CtsSinglePassageId identifies part of a range, or a non-range passage
    such as 1.1.1 or 1.2.2 in 1.2.2-1.2.3.

    It provides a list version of itself through the .list property and
    links to its subreference using the .subreference property (Word and Index identifier)

    If you iter over it, it returns each passage of the hierarchy, so

    >>> iter((CtsSinglePassageId("1.2.3"))) == iter(["1", "2", "3"])

    len() and .depth returns the depth of the passage

    >>> (CtsSinglePassageId("1.2.4")).depth == 3
    >>> len(CtsSinglePassageId("1.2.4")) == 3
    """
    def __new__(cls, str_repr: str):
        # Saving the original ID
        obj = str.__new__(cls, str_repr)

        # Setting up the properties that can be None
        obj._sub_reference = None

        # Parsing the reference
        temp_str_repr = str_repr
        subreference = temp_str_repr.split("@")

        if len(subreference) == 2:
            obj._sub_reference = CtsWordReference(subreference[1])
            temp_str_repr = subreference[0]

        obj._list = temp_str_repr
        return obj

    @property
    def list(self) -> List[str]:
        return list(iter(self))

    @property
    def subreference(self) -> Optional[CtsWordReference]:
        subref = self.split("@")
        if len(subref) == 2:
            return CtsWordReference(subref[1])

    def __iter__(self) -> List[str]:
        subref = self.split("@")[0]
        yield from subref.split(".")

    def __len__(self) -> int:
        return self.count(".") + 1

    @property
    def depth(self) -> int:
        return self.count(".") + 1


class CtsReference(BaseReference):
    """ A reference object giving information

    :Example:
        >>>    a = CtsReference("1.1@Achiles[1]-1.2@Zeus[1]")
        >>>    b = CtsReference("1.1")
        >>>    CtsReference("1.1-2.2.2").highest == CtsSinglePassageId("1.1")

    Reference object supports the following magic methods : len(), str() and eq().

    :Example:
        >>>    len(a) == 2 && len(b) == 1
        >>>    str(a) == "1.1@Achiles[1]-1.2@Zeus[1]"
        >>>    b == CtsReference("1.1") && b != a

    .. note::
        Reference(...).subreference and .list are not available for range. You will need to convert .start or .end to
        a Reference

        >>>    ref = CtsReference('1.2.3')
    """

    def __new__(cls, *references):
        # pickle.load will try to feed the tuple back !
        if len(references) == 2:
            start, end = references
            o = BaseReference.__new__(
                CtsReference,
                CtsSinglePassageId(start),
                CtsSinglePassageId(end)
            )
            o._str_repr = start + "-" + end
            return o

        references, *_ = references
        if not references:
            return None
        elif isinstance(references, tuple):
            if references[1]:
                o = BaseReference.__new__(
                    CtsReference,
                    CtsSinglePassageId(references[0]),
                    CtsSinglePassageId(references[1])
                )
            else:
                o = BaseReference.__new__(
                    CtsReference,
                    CtsSinglePassageId(references[0])
                )
            o._str_repr = "-".join([r for r in references if r])

        elif isinstance(references, str):
            if "-" not in references:
                o = BaseReference.__new__(CtsReference, CtsSinglePassageId(references))
            else:
                _start, _end = tuple(references.split("-"))
                o = BaseReference.__new__(CtsReference, CtsSinglePassageId(_start), CtsSinglePassageId(_end))
            o._str_repr = references

        return o

    @property
    def parent(self) -> Optional['CtsReference']:
        """ Parent of the actual URN, for example, 1.1 for 1.1.1

        :rtype: CtsReference
        """
        if self.start.depth == 1 and (self.end is None or self.end.depth <= 1):
            return None
        else:
            if self.start.depth > 1 and (self.end is None or self.end.depth == 0):
                return CtsReference("{0}{1}".format(
                    ".".join(self.start.list[:-1]),
                    self.start.subreference or ""
                ))
            elif self.start.depth > 1 and self.end is not None and self.end.depth > 1:
                _start = self.start.list[0:-1]
                _end = self.end.list[0:-1]
                if _start == _end and \
                        self.start.subreference is None and \
                        self.end.subreference is None:
                    return CtsReference(
                        ".".join(_start)
                    )
                else:
                    return CtsReference("{0}{1}-{2}{3}".format(
                        ".".join(_start),
                        self.start.subreference or "",
                        ".".join(_end),
                        self.end.subreference or ""
                    ))

    @property
    def highest(self) -> CtsSinglePassageId:
        """ Return highest reference level

        For references such as 1.1-1.2.8, with different level, it can be useful to access to the highest node in the
        hierarchy. In this case, the highest level would be 1.1. The function would return ["1", "1"]

        .. note:: By default, this property returns the start level

        :rtype: CtsReference
        """
        if not self.end:
            return self.start
        elif len(self.start) < len(self.end) and len(self.start):
            return self.start
        elif len(self.start) > len(self.end) and len(self.end):
            return self.end
        elif len(self.start):
            return self.start

    @property
    def start(self) -> CtsSinglePassageId:
        """ Quick access property for start list

        :rtype: str
        """
        return super(CtsReference, self).start

    @property
    def end(self) -> CtsSinglePassageId:
        """ Quick access property for reference end list

        :rtype: str
        """
        return super(CtsReference, self).end

    @property
    def subreference(self):
        """ Return the subreference of a single node reference

        .. note:: Access to start and end subreference should be done through obj.start.subreference \
        and obj.end.subreference

        :rtype: (str, int)
        """
        if not self.end:
            return self.start.subreference

    @property
    def depth(self):
        """ Return depth of highest reference level

        For references such as 1.1-1.2.8, or simple references such as 1.a, with different level, it can be useful to
        know the depth of the reference to access the right XPath for example. This property returns the depth of the
        highest node

        :example:
            - len(1.1) == 2
            - len(1.2.8-1.3) == 2
            - len(1-1.2) == 1

        :rtype: int
        """
        return len(self.highest.list)

    def __str__(self):
        """ Return full reference in string format

        :rtype: str
        :returns: String representation of Reference Object

        :Example:
            >>>    a = CtsReference("1.1@Achiles[1]-1.2@Zeus[1]")
            >>>    b = CtsReference("1.1")
            >>>    str(a) == "1.1@Achiles[1]-1.2@Zeus[1]"
            >>>    str(b) == "1.1"
        """
        return self._str_repr


class CtsReferenceSet(BaseReferenceSet):
    """ A CTS version of the BaseReferenceSet

    """
    def __contains__(self, item: str) -> bool:
        return BaseReferenceSet.__contains__(self, item) or \
               BaseReferenceSet.__contains__(self, CtsReference(item))

    def index(self, obj: Union[str, CtsReference], *args, **kwargs) -> int:
        _o = obj
        if not isinstance(obj, CtsReference):
            _o = CtsReference(obj)
        return super(CtsReferenceSet, self).index(_o)


class URN(object):
    """ A URN object giving all useful sections

    :param urn: A CTS URN
    :type urn: str
    :cvar NAMESPACE: Constant representing the URN until its namespace
    :cvar TEXTGROUP: Constant representing the URN until its textgroup
    :cvar WORK: Constant representing the URN until its work
    :cvar VERSION: Constant representing the URN until its version
    :cvar PASSAGE: Constant representing the URN until its full passage
    :cvar PASSAGE_START: Constant representing the URN until its passage (end excluded)
    :cvar PASSAGE_END: Constant representing the URN until its passage (start excluded)
    :cvar NO_PASSAGE: Constant representing the URN until its passage excluding its passage
    :cvar COMPLETE: Constant representing the complete URN

    :Example:
        >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")

    URN object supports the following magic methods : len(), str() and eq(), gt() and lt().

    :Example:
        >>>     b = URN("urn:cts:latinLit:phi1294.phi002")
        >>>     a != b
        >>>     a > b # It has more member. Only member count is compared
        >>>     b < a
        >>>     len(a) == 5 # CtsReference is not counted to not induce count equivalencies with the optional version
        >>>     len(b) == 4

    """

    NAMESPACE = 0
    TEXTGROUP = 1
    WORK = 2
    VERSION = 3
    PASSAGE = 4
    PASSAGE_START = 5
    PASSAGE_END = 6
    NO_PASSAGE = 10
    COMPLETE = 100

    def __init__(self, urn):
        self.__urn = None
        self.__parsed = self.__parse__(urn)

    @property
    def urn_namespace(self):
        """ General Namespace element of the URN

        :rtype: str
        :return: Namespace part of the URN
        """
        return self.__parsed["urn_namespace"]

    @urn_namespace.setter
    def urn_namespace(self, value):
        self.__urn = None
        self.__parsed["urn_namespace"] = value

    @property
    def namespace(self):
        """ CTS Namespace element of the URN

        :rtype: str
        :return: Namespace part of the URN
        """
        return self.__parsed["cts_namespace"]

    @namespace.setter
    def namespace(self, value):
        self.__urn = None
        self.__parsed["cts_namespace"] = value

    @property
    def textgroup(self):
        """ Textgroup element of the URN

        :rtype: str
        :return: Textgroup part of the URN
        """
        return self.__parsed["textgroup"]

    @textgroup.setter
    def textgroup(self, value):
        self.__urn = None
        self.__parsed["textgroup"] = value

    @property
    def work(self):
        """ CtsWorkMetadata element of the URN

        :rtype: str
        :return: CtsWorkMetadata part of the URN
        """
        return self.__parsed["work"]

    @work.setter
    def work(self, value):
        self.__urn = None
        self.__parsed["work"] = value

    @property
    def version(self):
        """ Version element of the URN

        :rtype: str
        :return: Version part of the URN
        """
        return self.__parsed["version"]

    @version.setter
    def version(self, value):
        self.__urn = None
        self.__parsed["version"] = value

    @property
    def reference(self):
        """ Reference element of the URN

        :rtype: CtsReference
        :return: Reference part of the URN
        """
        return self.__parsed["reference"]

    @reference.setter
    def reference(self, value):
        self.__urn = None
        if isinstance(value, CtsReference):
            self.__parsed["reference"] = value
        else:
            self.__parsed["reference"] = CtsReference(value)

    def __len__(self):
        """ Returns the len of the URN

        :rtype: int
        :returns: Length of the URN

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    print(len(a))
        """

        items = [
            key
            for key, value in self.__parsed.items()
            if key not in ["reference"] and value is not None
        ]
        return len(items)

    def __gt__(self, other):
        """ Allows for greater comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of bigger size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1")
            >>>    (a > b) == True
        """
        return len(self) > len(other)

    def __lt__(self, other):
        """ Allows for lower comparison

        :param other: Comparison object
        :type other: URN
        :rtype: boolean
        :returns: Indicator of lower size

        .. warning:: Does not take into account the passage !

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1")
            >>>    (b < a) == True
        """
        return len(self) < len(other)

    def __eq__(self, other):
        """ Equality checker for URN object

        :param other: An object to be checked against
        :type other: URN
        :rtype: boolean
        :returns: Equality between other and self

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    b = URN(urn="urn:cts:latinLit:phi1294.phi002:1.1")
            >>>    (b == a) == False
        """
        return isinstance(other, type(self)) and str(self) == str(other)

    def __ne__(self, other):
        """ Inequality checker for CtsReference object

        :param other: An object to be checked against
        :rtype: boolean
        :returns: Equality between other and self
        """
        return not self.__eq__(other)

    def __str__(self):
        """ Return full initial urn

        :rtype: basestring
        :returns: String representation of URN Object

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    str(a) == "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1"
        """
        if self.__urn is None:
            urn = "urn:" + self.__parsed["urn_namespace"]
            if self.namespace:
                urn += ":" + self.namespace
                if self.textgroup:
                    urn += ":" + self.textgroup
                    if self.work:
                        urn += "." + self.work
                        if self.version:
                            urn += "." + self.version
                        if self.reference:
                            urn += ":" + str(self.reference)
            self.__urn = urn
        return self.__urn

    def upTo(self, key):
        """ Returns the urn up to given level using URN Constants

        :param key: Identifier of the wished resource using URN constants
        :type key: int
        :returns: String representation of the partial URN requested
        :rtype: str

        :Example:
            >>>    a = URN(urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.1")
            >>>    a.upTo(URN.TEXTGROUP) == "urn:cts:latinLit:phi1294"
        """
        middle = [
            component
            for component in [self.__parsed["textgroup"], self.__parsed["work"], self.__parsed["version"]]
            if component is not None
        ]

        if key == URN.COMPLETE:
            return self.__str__()
        elif key == URN.NAMESPACE:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"]])
        elif key == URN.TEXTGROUP and self.__parsed["textgroup"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                self.__parsed["textgroup"]
            ])
        elif key == URN.WORK and self.__parsed["work"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join([self.__parsed["textgroup"], self.__parsed["work"]])
            ])
        elif key == URN.VERSION and self.__parsed["version"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle)
            ])
        elif key == URN.NO_PASSAGE and self.__parsed["work"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle)
            ])
        elif key == URN.PASSAGE and self.__parsed["reference"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference)
            ])
        elif key == URN.PASSAGE_START and self.__parsed["reference"]:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference.start)
            ])
        elif key == URN.PASSAGE_END and self.__parsed["reference"] and self.reference.end is not None:
            return ":".join([
                "urn",
                self.__parsed["urn_namespace"],
                self.__parsed["cts_namespace"],
                ".".join(middle),
                str(self.reference.end)
            ])
        else:
            raise KeyError("Provided key is not recognized.")

    @staticmethod
    def model():
        """ Generate a standard dictionary model for URN inside function

        :return: Dictionary of CTS elements
        """
        return {
            "urn_namespace": None,
            "cts_namespace": None,
            "textgroup": None,
            "work": None,
            "version": None,
            "reference": None
        }

    def __parse__(self, urn):
        """ Parse a URN

        :param urn: A URN:CTS
        :type urn: basestring
        :rtype: defaultdict.basestring
        :returns: Dictionary representation
        """
        parsed = URN.model()
        self.__urn = urn.split("#")[0]
        urn = self.__urn.split(":")
        if isinstance(urn, list) and len(urn) > 2:
            parsed["urn_namespace"] = urn[1]
            parsed["cts_namespace"] = urn[2]

            if len(urn) == 5:
                parsed["reference"] = CtsReference(urn[4])

            if len(urn) >= 4:
                urn = urn[3].split(".")
                if len(urn) >= 1:
                    parsed["textgroup"] = urn[0]
                if len(urn) >= 2:
                    parsed["work"] = urn[1]
                if len(urn) >= 3:
                    parsed["version"] = urn[2]
        else:
            raise ValueError("URN is empty")
        return parsed


class Citation(BaseCitation):
    """ A citation object gives informations about the scheme

    :param name: Name of the citation (e.g. "book")
    :type name: basestring
    :param xpath: Xpath of the citation (As described by CTS norm)
    :type xpath: basestring
    :param scope: Scope of the citation (As described by CTS norm)
    :type xpath: basestring
    :param refsDecl: refsDecl version
    :type refsDecl: basestring
    :param child: A citation
    :type child: XmlCtsCitation
    :ivar name: Name of the citation (e.g. "book")
    :type name: basestring
    :ivar xpath: Xpath of the citation (As described by CTS norm)
    :type xpath: basestring
    :ivar scope: Scope of the citation (As described by CTS norm)
    :type xpath: basestring
    :ivar refsDecl: refsDecl version
    :type refsDecl: basestring
    :ivar child: A citation
    :type child: Citation

    """

    EXPORT_TO = [Mimetypes.XML.CTS, Mimetypes.XML.TEI]
    DEFAULT_EXPORT = Mimetypes.XML.CTS

    escape = re.compile('(")')

    def __init__(self, name=None, xpath=None, scope=None, refsDecl=None, child=None):
        """ Initialize a XmlCtsCitation object
        """
        super(Citation, self).__init__(name=name, children=[child])

        self.__refsDecl = None

        self.name = name
        if scope and xpath:
            self._fromScopeXpathToRefsDecl(scope, xpath)
        else:
            self.refsDecl = refsDecl

    @property
    def xpath(self):
        """ CtsTextInventoryMetadata xpath property of a citation (ie. identifier of the last element of the citation)

        :type: basestring
        :Example: //tei:l[@n="?"]
        """
        return self._parseXpathScope()[1]

    @xpath.setter
    def xpath(self, new_xpath):
        if new_xpath is not None and self.refsDecl:
            current_scope, current_xpath = self._parseXpathScope()
            self._fromScopeXpathToRefsDecl(current_scope, new_xpath)

    @property
    def scope(self):
        """ CtsTextInventoryMetadata scope property of a citation (ie. identifier of all element but the last of the citation)

        :type: basestring
        :Example: /tei:TEI/tei:text/tei:body/tei:div
        """
        return self._parseXpathScope()[0]

    @scope.setter
    def scope(self, new_scope):
        if new_scope is not None and self.refsDecl:
            current_scope, current_xpath = self._parseXpathScope()
            self._fromScopeXpathToRefsDecl(new_scope, current_xpath)

    @property
    def refsDecl(self):
        """ ResfDecl expression of the citation scheme

        :rtype: str
        :Example: /tei:TEI/tei:text/tei:body/tei:div//tei:l[@n='$1']
        """
        return self.__refsDecl

    @refsDecl.setter
    def refsDecl(self, val):
        if val is not None:
            self.__refsDecl = val

    @property
    def child(self):
        """ Child of a citation

        :type: XmlCtsCitation or None
        :Example: XmlCtsCitation.name==poem would have a child XmlCtsCitation.name==line
        """
        if len(self.children):
            return self.children[0]

    @child.setter
    def child(self, val):
        if val:
            self.children = [val]
            if self.is_root():
                val.root = self
            else:
                val.root = self.root
        else:
            self.children = []

    @property
    def attribute(self):
        """ Attribute that serves as a reference getter
        """
        refs = re.findall(
            "\@([a-zA-Z:]+)=\\\?[\'\"]\$"+str(self.refsDecl.count("$"))+"\\\?[\'\"]",
            self.refsDecl
        )
        return refs[-1]

    def _parseXpathScope(self):
        """ Update xpath and scope property when refsDecl is updated

        :returns: Scope, Xpath
        """
        rd = self.refsDecl
        matches = REFSDECL_SPLITTER.findall(rd)
        return REFSDECL_REPLACER.sub("?", "".join(matches[0:-1])), REFSDECL_REPLACER.sub("?", matches[-1])

    def _fromScopeXpathToRefsDecl(self, scope, xpath):
        """ Update the refsDecl value if xpath and scope property are to be updated

        """
        if scope is not None and xpath is not None:
            _xpath = scope + xpath
            i = _xpath.find("?")
            ii = 1
            while i >= 0:
                _xpath = _xpath[:i] + "$" + str(ii) + _xpath[i+1:]
                i = _xpath.find("?")
                ii += 1
            self.refsDecl = _xpath

    def __getitem__(self, item):
        if not isinstance(item, int) or item > len(self)-1:
            raise KeyError("XmlCtsCitation index is too big")
        return [x for x in self][item]

    def __len__(self):
        """ Length method

        :rtype: int
        :returns: Number of nested citations
        """
        return len([x for x in self])

    def match(self, passageId):
        """ Given a passageId matches a citation level

        :param passageId: A passage to match
        :return:
        """
        if not isinstance(passageId, CtsReference):
            passageId = CtsReference(passageId)

        if self.is_root():
            return self[passageId.depth-1]
        return self.root.match(passageId)

    def fill(self, passage=None, xpath=None):
        """ Fill the xpath with given informations

        :param passage: CapitainsCtsPassage reference
        :type passage: CtsReference or list or None. Can be list of None and not None
        :param xpath: If set to True, will return the replaced self.xpath value and not the whole self.refsDecl
        :type xpath: Boolean
        :rtype: basestring
        :returns: Xpath to find the passage

        .. code-block:: python

            citation = XmlCtsCitation(name="line", scope="/TEI/text/body/div/div[@n=\"?\"]",xpath="//l[@n=\"?\"]")
            print(citation.fill(["1", None]))
            # /TEI/text/body/div/div[@n='1']//l[@n]
            print(citation.fill(None))
            # /TEI/text/body/div/div[@n]//l[@n]
            print(citation.fill(CtsReference("1.1"))
            # /TEI/text/body/div/div[@n='1']//l[@n='1']
            print(citation.fill("1", xpath=True)
            # //l[@n='1']


        """
        if xpath is True:  # Then passage is a string or None
            xpath = self.xpath

            replacement = r"\1"
            if isinstance(passage, str):
                replacement = r"\1\2'" + passage + "'"

            return REFERENCE_REPLACER.sub(replacement, xpath)
        else:
            if isinstance(passage, CtsReference):
                passage = passage.start.list
            elif passage is None:
                return REFERENCE_REPLACER.sub(
                    r"\1",
                    self.refsDecl
                )
            passage = iter(passage)
            return REFERENCE_REPLACER.sub(
                lambda m: _ref_replacer(m, passage),
                self.refsDecl
            )

    def is_set(self) -> bool:
        """ Check if the citation has been set

        :return: True if set up, False if not
        :rtype: bool
        """
        return self.refsDecl is not None

    def __export__(self, output=None, **kwargs):
        if output == Mimetypes.XML.CTS:
            if self.xpath is None and self.scope is None and self.refsDecl is None:
                return ""

            child = ""
            if isinstance(self.child, Citation):
                child = self.child.export(output=output)

            label = ""
            if self.name is not None:
                label = self.name

            return make_xml_node(
                get_graph(), RDF_NAMESPACES.CTS.citation, attributes={
                    "xpath": re.sub(Citation.escape, "'", self.xpath),
                    "scope": re.sub(Citation.escape, "'", self.scope),
                    "label": label
                }, innerXML=child, complete=True
            )
        elif output == Mimetypes.XML.TEI:
            if self.refsDecl is None:
                return ""

            label = ""
            if self.name is not None:
                label = self.name

            return \
                "<tei:cRefPattern n=\"{label}\" matchPattern=\"{regexp}\" replacementPattern=\"#xpath({refsDecl})\">" \
                "<tei:p>This pointer pattern extracts {label}</tei:p></tei:cRefPattern>".format(
                    refsDecl=self.refsDecl,
                    label=label,
                    regexp="\.".join(["(\w+)"]*self.refsDecl.count("$"))
                )

    def export(self, output=None, **kwargs):
        if self.refsDecl:
            return super(Citation, self).export(output=output, **kwargs)
        return ""

    @staticmethod
    def ingest(resource, xpath=".//tei:cRefPattern"):
        """ Ingest a resource and store data in its instance

        :param resource: XML node cRefPattern or list of them in ASC hierarchy order (deepest to highest, eg. lines to poem to book)
        :type resource: [lxml.etree._Element]
        :param xpath: XPath to use to retrieve citation
        :type xpath: str

        :returns: A citation object
        :rtype: Citation
        """
        if len(resource) == 0 and isinstance(resource, list):
            return None
        elif isinstance(resource, list):
            resource = resource[0]
        elif not isinstance(resource, _Element):
            return None

        resource = resource.xpath(xpath, namespaces=XPATH_NAMESPACES)
        citations = []

        for x in range(0, len(resource)):
            citations.append(
                Citation(
                    name=resource[x].get("n"),
                    refsDecl=resource[x].get("replacementPattern")[7:-1],
                    child=_child_or_none(citations)
                )
            )
        if len(citations) > 1:
            for citation in citations[:-1]:
                citation.root = citations[-1]
        return citations[-1]


def _ref_replacer(match, passage):
    """ Helper to replace xpath/scope/refsDecl on iteration with passage value

    :param match: A RegExp match
    :type match: re.SRE_MATCH
    :param passage: A list with subreference informations
    :type passage: iter

    :rtype: basestring
    :return: Replaced string
    """
    groups = match.groups()
    ref = next(passage)
    if ref is None:
        return groups[0]
    else:
        return "{1}='{0}'".format(ref, groups[0])


# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resolvers.cts.remote
   :synopsis: Resolver built for CTS APIs

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from typing import Union, Optional, Any, Dict
import re
from pyld.jsonld import expand

from MyCapytain.resolvers.prototypes import Resolver
from MyCapytain.common.reference import BaseReference, BaseReferenceSet, \
    DtsReference, DtsReferenceSet, DtsCitation
from MyCapytain.retrievers.dts import HttpDtsRetriever
from MyCapytain.common.utils.dts import parse_metadata
from MyCapytain.resources.collections.dts import HttpResolverDtsCollection
from MyCapytain.resources.texts.remote.dts import DtsResolverDocument
from MyCapytain.errors import EmptyReference, PaginationBrowsingError

__all__ = [
    "HttpDtsResolver"
]

_empty = [{"@value": None}]
_re_page = re.compile("page=(\d+)")


def _parse_ref(ref_dict, default_type: str=None):
    if "https://w3id.org/dts/api#ref" in ref_dict:
        refs = ref_dict["https://w3id.org/dts/api#ref"][0]["@value"],
    elif "https://w3id.org/dts/api#start" in ref_dict and \
         "https://w3id.org/dts/api#end" in ref_dict:
        refs = (
            ref_dict["https://w3id.org/dts/api#start"][0]["@value"],
            ref_dict["https://w3id.org/dts/api#end"][0]["@value"]
        )
    else:
        raise EmptyReference("A reference is either empty or malformed")
    type_ = default_type
    if "https://w3id.org/dts/api#citeType" in ref_dict:
        type_ = ref_dict["https://w3id.org/dts/api#citeType"][0]["@value"]

    obj = DtsReference(*refs, type_=type_)
    parse_metadata(obj.metadata, ref_dict)

    return obj


class HttpDtsResolver(Resolver):
    """ HttpDtsResolver provide a resolver for DTS API http endpoint.

    :param endpoint: DTS API Retriever
    """
    def __init__(self, endpoint: Union[str, HttpDtsRetriever]):
        if not isinstance(endpoint, HttpDtsRetriever):
            endpoint = HttpDtsRetriever(endpoint)
        self._endpoint = endpoint

    @property
    def endpoint(self) -> HttpDtsRetriever:
        """ DTS Endpoint of the resolver

        :return: DTS Endpoint
        :rtype: HttpDtsRetriever
        """
        return self._endpoint

    def getMetadata(self, objectId: str=None, **filters) -> HttpResolverDtsCollection:
        """ Retrieves metadata calling the Collections Endpoint """
        req = self.endpoint.get_collection(objectId)
        req.raise_for_status()

        collection = HttpResolverDtsCollection.parse(req.json(), resolver=self)
        # Pagination is not completed upon first query.
        # Pagination will be treated direction in the HttpResolverDtsCollection

        return collection

    def getReffs(
            self,
            textId: str,
            level: int=1,
            subreference: Union[str, BaseReference]=None,
            include_descendants: bool=False,
            additional_parameters: Optional[Dict[str, Any]]=None
    ) -> DtsReferenceSet:
        """ Retrieve references by calling the Navigation API """
        if not additional_parameters:
            additional_parameters = {}

        references = []
        default_type = None
        level_ = level

        page = 1
        while page:
            kwargs = dict(
                level=level, ref=subreference,
                exclude=additional_parameters.get("exclude", None),
                group_by=additional_parameters.get("groupBy", 1)
            )
            if page != 1:
                kwargs["page"] = page

            response = self.endpoint.get_navigation(textId, **kwargs)
            response.raise_for_status()

            data = response.json()
            data = expand(data)
            if not len(data):
                raise PaginationBrowsingError(
                    "The contacted endpoint seems to not have any data about collection %s " % self.id
                )
            data = data[0]

            level_ = data.get("https://w3id.org/dts/api#level", _empty)[0]["@value"]
            default_type = data.get("https://w3id.org/dts/api#citeType", _empty)[0]["@value"]
            members = data.get("https://www.w3.org/ns/hydra/core#member", [])

            references.extend([
                _parse_ref(ref, default_type=default_type)
                for ref in members
            ])

            page = None
            if "https://www.w3.org/ns/hydra/core#view" in data:
                if "https://www.w3.org/ns/hydra/core#next" in data["https://www.w3.org/ns/hydra/core#view"][0]:
                    page = _re_page.findall(
                        data["https://www.w3.org/ns/hydra/core#view"]
                            [0]["https://www.w3.org/ns/hydra/core#next"]
                            [0]["@value"]
                    )[0]

        citation = None
        if default_type:
            citation = DtsCitation(name=default_type)

        reffs = DtsReferenceSet(
            *references,
            level=level_,
            citation=citation
        )
        return reffs

    def getTextualNode(
            self,
            textId: str,
            subreference: Union[str, BaseReference]=None,
            prevnext: bool=False,
            metadata: bool=False
    ) -> DtsResolverDocument:
        """ Retrieve a text node from the API via the Document Endpoint

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :param prevnext: Retrieve graph representing previous and next passage
        :type prevnext: boolean
        :param metadata: Retrieve metadata about the passage and the text
        :type metadata: boolean
        :return: CapitainsCtsPassage
        :rtype: CapitainsCtsPassage
        """
        return DtsResolverDocument.parse(
            identifier=textId,
            reference=subreference,
            resolver=self,
            response=self.endpoint.get_document(collection_id=textId, ref=subreference)
        )



from ._json_ld import dict_to_literal
from rdflib import URIRef
from ..metadata import Metadata


__all__ = [
    "parse_metadata"
]


def parse_metadata(metadata_obj: Metadata, metadata_dictionary: dict) -> None:
    """ Adds to a Metadata object any DublinCore or dts:Extensions object
    found in the given dictionary

    :param metadata_obj:
    :param metadata_dictionary:
    """
    for key, value_set in metadata_dictionary.get("https://w3id.org/dts/api#dublincore", [{}])[0].items():
        term = URIRef(key)
        for value_dict in value_set:
            metadata_obj.add(term, *dict_to_literal(value_dict))

    for key, value_set in metadata_dictionary.get("https://w3id.org/dts/api#extensions", [{}])[0].items():
        term = URIRef(key)
        for value_dict in value_set:
            metadata_obj.add(term, *dict_to_literal(value_dict))



from typing import TYPE_CHECKING, Optional
import link_header
from urllib.parse import parse_qs, urlparse

from MyCapytain.common.reference import BaseReference, DtsReference
from MyCapytain.common.utils.xml import xmlparser
from MyCapytain.resources.texts.base.tei import TeiResource


if TYPE_CHECKING:  # Required for nice and beautiful type checking
    from requests import Response
    from lxml.etree import ElementTree
    from MyCapytain.resolvers.dts.api_v1 import HttpDtsResolver
    from MyCapytain.resources.collections.dts import HttpResolverDtsCollection


class DtsResolverDocument(TeiResource):
    def __init__(
            self,
            identifier: str,
            resolver: "HttpDtsResolver",
            resource: "ElementTree",
            reference: Optional[DtsReference]=None,
            collection: Optional["HttpResolverDtsCollection"]=None,
            **kwargs
    ):
        super(DtsResolverDocument, self).__init__(identifier=identifier, resource=resource, **kwargs)
        self._resolver = resolver
        self._reference = reference
        self._resource = resource

        self._parent = None
        self._prev_id, self._next_id = None, None
        self._parsed_metadata = False
        self._collection = collection
        self._depth = None
        self._metadata = None

    @property
    def collection(self):
        if not self._collection:
            self._collection = self.resolver.getMetadata(self.id)
        return self._collection

    @property
    def depth(self):
        if self._depth is None:
            self._depth = self.collection.citation.depth
        return self._depth

    @property
    def metadata(self):
        if not self._metadata:
            self._metadata = self.collection.metadata
        return self._metadata

    @property
    def resolver(self) -> "HttpDtsResolver":
        return self._resolver

    def getTextualNode(self, subreference: DtsReference) -> "DtsResolverPassage":
        obj = self.resolver.getTextualNode(textId=self.id, subreference=subreference)
        # We set up the collection with the current one :
        #   if it has been retrieved, it'll be used
        obj._collection = self._collection
        return obj

    def getReffs(self, level: int=1, subreference: DtsReference=None):
        return self.resolver.getReffs(textId=self.id, subreference=(subreference or self.reference))

    def _dict_to_ref(self, header_dic):
        ref = header_dic.get("ref", [None])[0]
        if ref:
            return DtsReference(ref)

        s = header_dic.get("start", [None])[0]
        e = header_dic.get("end", [None])[0]
        if s and e:
            return DtsReference(s, e)

    @classmethod
    def parse(cls, identifier: str, reference: DtsReference, resolver: "HttpDtsResolver", response: "Response"):
        o = cls(
            identifier=identifier,
            reference=reference,
            resolver=resolver,
            resource=xmlparser(response.text)
        )

        links = link_header.parse(response.headers.get("Link", ""))
        links = {
            link.rel: parse_qs(urlparse(link.href).query)
            for link in links.links
        }
        if links.get("next"):
            o._next_id = o._dict_to_ref(links.get("next"))
        if links.get("prev"):
            o._prev_id = o._dict_to_ref(links.get("prev"))
        if links.get("parent"):
            o._parent = o._dict_to_ref(links.get("up"))
        if links.get("first"):
            o._first_id = o._dict_to_ref(links.get("first"))
        if links.get("parent"):
            o._last_id = o._dict_to_ref(links.get("last"))
        if links.get("collection"):
            o._collection = o._dict_to_ref(links.get("collection"))

        return o

    @property
    def reference(self) -> DtsReference:
        return self._reference



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.common.reference
   :synopsis: URN related objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

"""
from ._base import NodeId, BaseCitationSet, BaseReference, BaseReferenceSet
from ._capitains_cts import Citation, CtsReference, CtsReferenceSet, URN
from ._dts_1 import DtsCitation, DtsCitationSet, DtsReference, DtsReferenceSet



from rdflib import Namespace, Graph
from rdflib.namespace import SKOS


__all__ = [
    "XPATH_NAMESPACES",
    "RDF_NAMESPACES",
    "Mimetypes",
    "GRAPH_BINDINGS",
    "bind_graph",
    "get_graph",
    "set_graph",
    "RDFLIB_MAPPING"
]

#: List of XPath Namespaces used in guidelines
XPATH_NAMESPACES = {
    "tei": "http://www.tei-c.org/ns/1.0",
    "ti": "http://chs.harvard.edu/xmlns/cts",
    "cpt": "http://purl.org/capitains/ns/1.0#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",
    "xml": "http://www.w3.org/XML/1998/namespace",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
}


class RDF_NAMESPACES:
    """ Namespaces Constants used to provide Namespace capacities across the library

    :cvar CTS: CTS Namespace
    :type CTS: Namespace
    :cvar TEI: TEI Namespace
    :type TEI: Namespace
    :cvar DC: DC Elements
    :type DC: Namespace
    :cvar CAPITAINS: CapiTainS Ontology
    :type CAPITAINS: Namespace
    """
    CTS = Namespace("http://chs.harvard.edu/xmlns/cts/")
    DTS = Namespace("https://w3id.org/dts/api#")
    TEI = Namespace("http://www.tei-c.org/ns/1.0/")
    CAPITAINS = Namespace("http://purl.org/capitains/ns/1.0#")
    HYDRA = Namespace("https://www.w3.org/ns/hydra/core#")


class Mimetypes:
    """ Mimetypes constants that are used to provide export functionality to base MyCapytain object.

    :cvar JSON: JSON Resource mimetype
    :cvar XML: XML Resource mimetype
    :cvar PYTHON: Python Native Object
    :cvar PLAINTEXT: Plain string format

    """

    class JSON:
        """ Json Mimetype

        :cvar Std: Standard JSON Export
        :cvar CTS: CTS Json Export
        """
        Std = "application/text"
        CTS = "application/ld+json:CTS"
        LD = "application/ld+json"

        class DTS:
            """ JSON DTS Expression

            :cvar Std: Standard DTS Json-LD Expression
            :cvar NoParents: DTS Json-LD Expression without parents expression
            """
            Std = "application/ld+json:DTS"
            NoParents = "application/ld+json:DTS/NoParents"

    class XML:
        """ XML Mimetype

        :cvar Std: Standard XML Export
        :cvar RDF: RDF XML Expression Export
        :cvar CTS: CTS API XML Expression Export
        :cvar TEI: TEI XML Expression Export
        """
        Std = "text/xml"
        RDF = "application/rdf+xml"
        RDFa = "application/rdfa+xml"
        CTS = "text/xml:CTS"
        TEI = "text/xml:tei"

        class CapiTainS:
            """ CapiTainS Guideline XML Structured metadata

            """
            CTS = "text/xml:CTS_CapiTainS"

    class PYTHON:
        """ Python Native Objects

        :cvar NestedDict: Nested Dictionary Object
        :cvar ETREE: Python LXML Etree Object
        """
        NestedDict = "python/NestedDict"
        ETREE = "python/lxml"

        class MyCapytain:
            """ MyCapytain Objects

            :cvar ReadableText: MyCapytain.resources.prototypes.text.CtsText
            """
            TextualElement = "Capitains/TextualElement"

    PLAINTEXT = "text/plain"


GRAPH_BINDINGS = {
    "": RDF_NAMESPACES.CTS,
    "dts": RDF_NAMESPACES.DTS,
    "tei": RDF_NAMESPACES.TEI,
    "skos": SKOS,
    "cpt": RDF_NAMESPACES.CAPITAINS
}


def bind_graph(graph=None):
    """ Bind a graph with generic MyCapytain prefixes

    :param graph: Graph (Optional)
    :return: Bound graph
    """
    if graph is None:
        graph = Graph()
    for prefix, ns in GRAPH_BINDINGS.items():
        graph.bind(prefix, ns, True)
    return graph


global __MYCAPYTAIN_TRIPLE_GRAPH__
__MYCAPYTAIN_TRIPLE_GRAPH__ = bind_graph()


def set_graph(graph):
    global __MYCAPYTAIN_TRIPLE_GRAPH__
    __MYCAPYTAIN_TRIPLE_GRAPH__ = graph


def get_graph():
    return __MYCAPYTAIN_TRIPLE_GRAPH__

RDFLIB_MAPPING = {
    Mimetypes.XML.RDF: "xml",
    Mimetypes.JSON.LD: "json-ld",
    Mimetypes.JSON.DTS.Std: "json-ld"
}




# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml
   :synopsis: XML based CtsTextMetadata and repository

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""
from __future__ import unicode_literals

from rdflib import URIRef, Literal
from rdflib.namespace import XSD
from lxml.objectify import IntElement, FloatElement

from MyCapytain.resources.prototypes.cts import inventory as cts
from MyCapytain.common.reference._capitains_cts import Citation as CitationPrototype
from MyCapytain.common.utils import expand_namespace
from MyCapytain.common.utils.xml import xmlparser
from MyCapytain.common.constants import XPATH_NAMESPACES, Mimetypes, RDF_NAMESPACES


__all__ = [
    "XmlCtsCitation",
    "XmlCtsWorkMetadata",
    "XmlCtsCommentaryMetadata",
    "XmlCtsTranslationMetadata",
    "XmlCtsEditionMetadata",
    "XmlCtsTextgroupMetadata",
    "XmlCtsTextInventoryMetadata",
    "XmlCtsTextMetadata"
]

_CLASSES_DICT = {}


class XmlCtsCitation(CitationPrototype):
    """ XmlCtsCitation XML implementation for CtsTextInventoryMetadata

    """

    @classmethod
    def ingest(cls, resource, element=None, xpath="ti:citation"):
        """ Ingest xml to create a citation

        :param resource: XML on which to do xpath
        :param element: Element where the citation should be stored
        :param xpath: XPath to use to retrieve citation

        :return: XmlCtsCitation
        """
        # Reuse of of find citation
        results = resource.xpath(xpath, namespaces=XPATH_NAMESPACES)
        if len(results) > 0:
            citation = cls(
                name=results[0].get("label"),
                xpath=results[0].get("xpath"),
                scope=results[0].get("scope")
            )

            if isinstance(element, cls):
                element.child = citation
                cls.ingest(
                    resource=results[0],
                    element=element.child
                )
            else:
                element = citation
                cls.ingest(
                    resource=results[0],
                    element=element
                )

            return citation

        return None


def _xpathDict(xml, xpath, cls, parent, **kwargs):
    """ Returns a default Dict given certain information

    :param xml: An xml tree
    :type xml: etree
    :param xpath: XPath to find children
    :type xpath: str
    :param cls: Class identifying children
    :type cls: inventory.Resource
    :param parent: Parent of object
    :type parent: CtsCollection
    :rtype: collections.defaultdict.<basestring, inventory.Resource>
    :returns: Dictionary of children
    """
    children = []
    for child in xml.xpath(xpath, namespaces=XPATH_NAMESPACES):
        children.append(cls.parse(
            resource=child,
            parent=parent,
            **kwargs
        ))
    return children


def _parse_structured_metadata(obj, xml):
    """ Parse an XML object for structured metadata

    :param obj: Object whose metadata are parsed
    :param xml: XML that needs to be parsed
    """
    for metadata in xml.xpath("cpt:structured-metadata/*", namespaces=XPATH_NAMESPACES):
        tag = metadata.tag
        if "{" in tag:
            ns, tag = tuple(tag.split("}"))
            tag = URIRef(ns[1:]+tag)
            s_m = str(metadata)
            if s_m.startswith("urn:") or s_m.startswith("http:") or s_m.startswith("https:") or s_m.startswith("hdl:"):
                obj.metadata.add(
                    tag,
                    URIRef(metadata)
                )
            elif '{http://www.w3.org/XML/1998/namespace}lang' in metadata.attrib:
                obj.metadata.add(
                    tag,
                    s_m,
                    lang=metadata.attrib['{http://www.w3.org/XML/1998/namespace}lang']
                )
            else:
                if "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}datatype" in metadata.attrib:
                    datatype = metadata.attrib["{http://www.w3.org/1999/02/22-rdf-syntax-ns#}datatype"]
                    if not datatype.startswith("http") and ":" in datatype:
                        datatype = expand_namespace(metadata.nsmap, datatype)
                    obj.metadata.add(tag, Literal(s_m, datatype=URIRef(datatype)))
                elif isinstance(metadata, IntElement):
                    obj.metadata.add(tag, Literal(int(metadata), datatype=XSD.integer))
                elif isinstance(metadata, FloatElement):
                    obj.metadata.add(tag, Literal(float(metadata), datatype=XSD.float))
                else:
                    obj.metadata.add(tag, s_m)


class XmlCtsTextMetadata(cts.CtsTextMetadata):
    """ Represents a CTS CtsTextMetadata

    """
    DEFAULT_EXPORT = Mimetypes.PYTHON.ETREE
    CLASS_CITATION = XmlCtsCitation

    @staticmethod
    def __findCitations(obj, xml, xpath="ti:citation"):
        """ Find citation in current xml. Used as a loop for xmlparser()

        :param xml: Xml resource to be parsed
        :param xpath: Xpath to use to retrieve the xml node
        """

    @classmethod
    def parse_metadata(cls, obj, xml):
        """ Parse a resource to feed the object

        :param obj: Obj to set metadata of
        :type obj: XmlCtsTextMetadata
        :param xml: An xml representation object
        :type xml: lxml.etree._Element
        """

        for child in xml.xpath("ti:description", namespaces=XPATH_NAMESPACES):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                obj.set_cts_property("description", child.text, lg)

        for child in xml.xpath("ti:label", namespaces=XPATH_NAMESPACES):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                obj.set_cts_property("label", child.text, lg)

        obj.citation = cls.CLASS_CITATION.ingest(xml, obj.citation, "ti:online/ti:citationMapping/ti:citation")

        # Added for commentary
        for child in xml.xpath("ti:about", namespaces=XPATH_NAMESPACES):
            obj.set_link(RDF_NAMESPACES.CTS.term("about"), child.get('urn'))

        _parse_structured_metadata(obj, xml)

        """
        online = xml.xpath("ti:online", namespaces=NS)
        if len(online) > 0:
            online = online[0]
            obj.docname = online.get("docname")
            for validate in online.xpath("ti:validate", namespaces=NS):
                obj.validate = validate.get("schema")
            for namespaceMapping in online.xpath("ti:namespaceMapping", namespaces=NS):
                obj.metadata["namespaceMapping"][namespaceMapping.get("abbreviation")] = namespaceMapping.get("nsURI")
        """

    def __init__(self, *args, **kwargs):
        super(XmlCtsTextMetadata, self).__init__(*args, **kwargs)
        self._path = None

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value


class XmlCtsEditionMetadata(cts.CtsEditionMetadata, XmlCtsTextMetadata):
    """ Create an edition subtyped CtsTextMetadata object
    """
    @classmethod
    def parse(cls, resource, parent=None):
        xml = xmlparser(resource)
        o = cls(urn=xml.get("urn"), parent=parent)
        cls.parse_metadata(o, xml)
        return o


class XmlCtsTranslationMetadata(cts.CtsTranslationMetadata, XmlCtsTextMetadata):
    """ Create a translation subtyped CtsTextMetadata object
    """
    @classmethod
    def parse(cls, resource, parent=None):
        xml = xmlparser(resource)
        lang = xml.get("{http://www.w3.org/XML/1998/namespace}lang")

        o = cls(urn=xml.get("urn"), parent=parent)
        if lang is not None:
            o.lang = lang
        cls.parse_metadata(o, xml)
        return o


class XmlCtsCommentaryMetadata(cts.CtsCommentaryMetadata, XmlCtsTextMetadata):
    """ Create a commentary subtyped PrototypeText object
    """
    @classmethod
    def parse(cls, resource, parent=None):
        xml = xmlparser(resource)
        lang = xml.get("{http://www.w3.org/XML/1998/namespace}lang")

        o = cls(urn=xml.get("urn"), parent=parent)
        if lang is not None:
            o.lang = lang
        cls.parse_metadata(o, xml)
        return o


class XmlCtsWorkMetadata(cts.CtsWorkMetadata):
    """ Represents a CTS Textgroup in XML
    """
    CLASS_EDITION = XmlCtsEditionMetadata
    CLASS_TRANSLATION = XmlCtsTranslationMetadata
    CLASS_COMMENTARY = XmlCtsCommentaryMetadata

    @classmethod
    def parse(cls, resource, parent=None, _with_children=False):
        """ Parse a resource

        :param resource: Element rerpresenting a work
        :param parent: Parent of the object
        :type parent: XmlCtsTextgroupMetadata
        """
        xml = xmlparser(resource)
        o = cls(urn=xml.get("urn"), parent=parent)

        lang = xml.get("{http://www.w3.org/XML/1998/namespace}lang")
        if lang is not None:
            o.lang = lang

        for child in xml.xpath("ti:title", namespaces=XPATH_NAMESPACES):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                o.set_cts_property("title", child.text, lg)

        # Parse children
        children = []
        children.extend(_xpathDict(
            xml=xml, xpath='ti:edition',
            cls=cls.CLASS_EDITION, parent=o
        ))
        children.extend(_xpathDict(
            xml=xml, xpath='ti:translation',
            cls=cls.CLASS_TRANSLATION, parent=o
        ))
        children.extend(_xpathDict(
            xml=xml, xpath='ti:commentary',
            cls=cls.CLASS_COMMENTARY, parent=o
        ))

        _parse_structured_metadata(o, xml)

        if _with_children:
            return o, children
        return o


class XmlCtsTextgroupMetadata(cts.CtsTextgroupMetadata):
    """ Represents a CTS Textgroup in XML
    """
    CLASS_WORK = XmlCtsWorkMetadata

    @classmethod
    def parse(cls, resource, parent=None):
        """ Parse a textgroup resource

        :param resource: Element representing the textgroup
        :param parent: Parent of the textgroup
        """
        xml = xmlparser(resource)
        o = cls(urn=xml.get("urn"), parent=parent)

        for child in xml.xpath("ti:groupname", namespaces=XPATH_NAMESPACES):
            lg = child.get("{http://www.w3.org/XML/1998/namespace}lang")
            if lg is not None:
                o.set_cts_property("groupname", child.text, lg)

        # Parse Works
        _xpathDict(xml=xml, xpath='ti:work', cls=cls.CLASS_WORK, parent=o)

        _parse_structured_metadata(o, xml)
        return o


class XmlCtsTextInventoryMetadata(cts.CtsTextInventoryMetadata):
    """ Represents a CTS Inventory file
    """
    CLASS_TEXTGROUP = XmlCtsTextgroupMetadata

    @classmethod
    def parse(cls, resource):
        """ Parse a resource

        :param resource: Element representing the text inventory
        """
        xml = xmlparser(resource)
        o = cls(name=xml.xpath("//ti:TextInventory", namespaces=XPATH_NAMESPACES)[0].get("tiid") or "")
        # Parse textgroups
        _xpathDict(xml=xml, xpath='//ti:textgroup', cls=cls.CLASS_TEXTGROUP, parent=o)
        return o



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.xml

Shared elements for TEI XmlCtsCitation

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>
"""
from lxml.etree import tostring

from MyCapytain.common.constants import Mimetypes, XPATH_NAMESPACES
from MyCapytain.common.utils import nested_ordered_dictionary, nested_set, normalize
from MyCapytain.common.utils.xml import xmlparser
from MyCapytain.resources.prototypes.text import InteractiveTextualNode


__all__ = [
    "TeiResource"
]


class TeiResource(InteractiveTextualNode):
    """ TEI Encoded Resource

    :param resource: XML Resource that needs to be parsed into a CapitainsCtsPassage/CtsTextMetadata
    :type resource: Union[str,_Element]
    :cvar EXPORT_TO: List of exportable supported formats
    :cvar DEFAULT_EXPORT: Default export (Plain/CtsTextMetadata)
    """
    EXPORT_TO = [
        Mimetypes.PYTHON.ETREE, Mimetypes.XML.Std,
        Mimetypes.PYTHON.NestedDict, Mimetypes.PLAINTEXT, Mimetypes.XML.TEI
    ]
    DEFAULT_EXPORT = Mimetypes.PLAINTEXT
    PLAINTEXT_STRING_JOIN = " "

    def __init__(self, resource: str, **kwargs):
        super(TeiResource, self).__init__(**kwargs)
        self.resource = xmlparser(resource)
        self._plaintext_string_join = "" + self.PLAINTEXT_STRING_JOIN

    @property
    def plaintext_string_join(self) -> str:
        """ String used to join xml node's texts in export
        """
        return self._plaintext_string_join

    @plaintext_string_join.setter
    def plaintext_string_join(self, value):
        """ Set the value for string join

        :param value: Default string value to use for join at export for plaintext
        :type value: str
        :return:
        """
        self._plaintext_string_join = value

    def __str__(self):
        """ CtsTextMetadata based representation of the passage

        :rtype: basestring
        :returns: XML of the passage in string form
        """
        return self.export(output=Mimetypes.XML.Std)

    def __export__(self, output=Mimetypes.PLAINTEXT, exclude=None, _preformatted=False):
        """ CtsTextMetadata content of the passage

        :param output: Mimetype (From MyCapytian.common.utils.Mimetypes) to output
        :type output: str
        :param exclude: Remove some nodes from text
        :type exclude: list
        :param _preformatted: This parameter is used when export loops on itself
        :type _preformatted: boolean
        :rtype: basestring
        :returns: CtsTextMetadata of the xml node

        :Example:
            >>>    P = CapitainsCtsPassage(resource='<l n="8">Ibis <note>hello<a>b</a></note> ab excusso missus in astra <hi>sago.</hi> </l>')
            >>>    P.export(output=Mimetypes.PLAINTEXT) == "Ibis hello b ab excusso missus in astra sago. "
            >>>    P.export(output=Mimetypes.PLAINTEXT, exclude=[]) == "Ibis hello b ab excusso missus in astra sago. "


        """
        if exclude is None:
            exclude = self.default_exclude

        if len(exclude) > 0 and _preformatted is False:
            exclude = "[{0}]".format(
                " and ".join(
                    "not(./ancestor-or-self::{0})".format(excluded)
                    for excluded in exclude
                )
            )
        elif _preformatted is False:
            exclude = ""

        if output == Mimetypes.PYTHON.ETREE:
            # Exports the whole resource as a LXML object
            return self.resource

        elif output == Mimetypes.XML.Std or output == Mimetypes.XML.TEI:
            # Exports the whole resource formatted as XML but as string object
            return tostring(self.resource, encoding=str)

        elif output == Mimetypes.PYTHON.NestedDict:
            # Exports the whole resource into a NestedDict
            reffs = self.getReffs(level=len(self.citation))
            text = nested_ordered_dictionary()
            for reff in reffs:
                _r = str(reff).split(".")  # Only works for non range of course
                nested_set(text, _r, self.getTextualNode(_r).export(
                    Mimetypes.PLAINTEXT,
                    exclude=exclude,
                    _preformatted=True
                ))
            return text

        elif output == Mimetypes.PLAINTEXT:
            # Exports to string
            return normalize(
                self.plaintext_string_join.join(
                    [
                        element
                        for element
                        in self.resource.xpath(
                            ".//descendant-or-self::text(){}".format(exclude),
                            namespaces=XPATH_NAMESPACES,
                            smart_strings=False
                        )
                    ]
                )
            )

    @property
    def xml(self):
        """ XML Representation of the CapitainsCtsPassage

        :rtype: lxml.etree._Element
        :returns: XML element representing the passage
        """
        return self.resource


from rdflib import Literal, URIRef


def literal_to_dict(value):
    """ Transform an object value into a dict readable value

    :param value: Object of a triple which is not a BNode
    :type value: Literal or URIRef
    :return: dict or str or list
    """
    if isinstance(value, Literal):
        if value.language is not None:
            return {"@value": str(value), "@language": value.language}
        return value.toPython()
    elif isinstance(value, URIRef):
        return {"@id": str(value)}
    elif value is None:
        return None
    return str(value)


def dict_to_literal(dict_container: dict):
    """ Transforms a JSON+LD PyLD dictionary into
    an RDFLib object"""
    if isinstance(dict_container["@value"], int):
        return dict_container["@value"],
    else:
        return dict_container["@value"], dict_container.get("@language", None)